#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a low pressure pump unit
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_pt, ZeroOrderBaseData
from idaes.core.util.constants import Constants

# Some more information about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("PumpElectricityZO")
class PumpElectricityZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a low pressure pump unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "pump_electricity"

        build_pt(self)

        self.lift_height = Var(units=pyunits.m, doc="Lift height for pump")

        self.eta_pump = Var(units=pyunits.dimensionless, doc="Efficiency of pump")

        self.eta_motor = Var(units=pyunits.dimensionless, doc="Efficiency of motor")

        self.electricity = Var(
            self.flowsheet().config.time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Electricity for low pressure pump",
        )

        self.applied_pressure = Var(
            self.flowsheet().config.time,
            units=pyunits.bar,
            bounds=(0, None),
            doc="Applied pressure",
        )

        self._fixed_perf_vars.append(self.lift_height)
        self._fixed_perf_vars.append(self.eta_pump)
        self._fixed_perf_vars.append(self.eta_motor)

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for electricity consumption based on " "pump flowrate.",
        )
        def electricity_consumption(b, t):
            return b.electricity[t] == pyunits.convert(
                b.lift_height
                * b.properties[t].flow_vol
                * b.properties[t].dens_mass
                * Constants.acceleration_gravity
                / (b.eta_pump * b.eta_motor),
                to_units=pyunits.kW,
            )

        @self.Constraint(
            self.flowsheet().time, doc="Constraint for pump applied pressure"
        )
        def applied_pressure_constraint(b, t):
            return b.applied_pressure[t] == pyunits.convert(
                b.lift_height
                * b.properties[t].dens_mass
                * Constants.acceleration_gravity,
                to_units=pyunits.bar,
            )

        self._perf_var_dict["Electricity (kW)"] = self.electricity
        self._perf_var_dict["Applied Pressure (bar)"] = self.applied_pressure

    @property
    def default_costing_method(self):
        return self.cost_pump_electricity

    @staticmethod
    def cost_pump_electricity(blk):
        """
        General method for costing low pressure pump. Capital cost
        is based on the cost of inlet flow rate.
        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A = blk.unit_model._get_tech_parameters(
            blk, parameter_dict, blk.unit_model.config.process_subtype, ["pump_cost"]
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = pyo.units.convert(
            blk.unit_model.properties[t0].flow_vol * A,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        # Determine if a costing factor is required
        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import os, pathlib, warnings
import h5py
import itertools
import pprint
import copy
import numpy as np

from scipy.interpolate import griddata

from pyomo.common.config import ConfigDict, ConfigValue


class ParameterSweepWriter:
    CONFIG = ConfigDict()

    CONFIG.declare(
        "debugging_data_dir",
        ConfigValue(
            default=None,
            domain=str,
            description="directory path to output debugging data.",
        ),
    )

    CONFIG.declare(
        "csv_results_file_name",
        ConfigValue(
            default=None, domain=str, description="filepath to the output CSV file."
        ),
    )

    CONFIG.declare(
        "h5_results_file_name",
        ConfigValue(
            default=None, domain=str, description="filepath to the output H5 file."
        ),
    )

    CONFIG.declare(
        "interpolate_nan_outputs",
        ConfigValue(
            default=False,
            domain=bool,
            description="Bool to decide whether to interpolate NaN outputs.",
        ),
    )

    CONFIG.declare(
        "h5_parent_group_name",
        ConfigValue(
            default=None,
            domain=str,
            description="Parent group name (container like objects, similar to a folder/directory in a file system) for parameter sweep outputs to be saved.",
        ),
    )

    def __init__(
        self,
        parallel_manager,
        **options,
    ):
        self.parallel_manager = parallel_manager
        self.config = self.CONFIG(options)

        if self.parallel_manager.is_root_process():
            if (
                self.config.h5_results_file_name is None
                and self.config.csv_results_file_name is None
            ):
                warnings.warn(
                    "No results will be writen to disk as h5_results_file_name and csv_results_file_name are both None"
                )

    @staticmethod
    def _strip_extension(file_name, extension):
        if file_name.lower().endswith(extension):
            return file_name[: -len(extension)], extension
        else:
            return file_name, None

    def _process_results_filename(self, results_file_name):
        # Get the directory path
        dirname = os.path.dirname(results_file_name)
        # Get the file name without the extension
        known_extensions = [".h5", ".csv"]
        for ext in known_extensions:
            fname_no_ext, extension = self._strip_extension(results_file_name, ext)
            if extension is not None:
                break

        return dirname, fname_no_ext, extension

    @staticmethod
    def _interp_nan_values(global_values, global_results):
        global_results_clean = np.copy(global_results)

        n_vals = np.shape(global_values)[1]
        n_outs = np.shape(global_results)[1]

        # Build a mask of all the non-nan saved outputs
        # i.e., where the optimzation succeeded
        mask = np.isfinite(global_results[:, 0])

        # Create a list of points where good data is available
        x0 = global_values[mask, :]

        if np.sum(mask) >= 4:
            # Interpolate to get a value for nan points where possible
            for k in range(n_outs):
                y0 = global_results[mask, k]
                yi = griddata(
                    x0, y0, global_values, method="linear", rescale=True
                ).reshape(-1)
                global_results_clean[~mask, k] = yi[~mask]

        else:
            warnings.warn("Too few points to perform interpolation.")

        return global_results_clean

    def _write_debug_data(
        self,
        sweep_params,
        local_values,
        local_results_dict,
        write_h5,
        write_csv,
        process_number,
    ):
        if write_h5:
            fname_h5 = f"local_results_{process_number:03}.h5"
            self._write_output_to_h5(
                local_results_dict,
                os.path.join(self.config["debugging_data_dir"], fname_h5),
            )
        if write_csv:
            fname_csv = f"local_results_{process_number:03}.csv"

            data_header = ",".join(itertools.chain(sweep_params))
            local_results = np.zeros(
                (np.shape(local_values)[0], len(local_results_dict["outputs"])),
                dtype=float,
            )
            for i, (key, item) in enumerate(local_results_dict["outputs"].items()):
                data_header = ",".join([data_header, key])
                local_results[:, i] = item["value"][:]

            local_save_data = np.hstack((local_values, local_results))

            # Save the local data
            np.savetxt(
                os.path.join(self.config["debugging_data_dir"], fname_csv),
                local_save_data,
                header=data_header,
                delimiter=", ",
                fmt="%.6e",
            )

    def _write_outputs(self, output_dict, txt_options="metadata"):
        self._write_output_to_h5(output_dict, self.config["h5_results_file_name"])

        # We will also create a companion txt file by default which contains
        # the metadata of the h5 file in a user readable format.
        txt_fname = self.config["h5_results_file_name"] + ".txt"

        if (
            self.config.h5_parent_group_name is None
            and os.path.isfile(txt_fname) is False
        ):
            if txt_options == "metadata":
                my_dict = copy.deepcopy(output_dict)
                for key, value in my_dict.items():
                    for subkey, subvalue in value.items():
                        subvalue.pop("value")
            elif txt_options == "keys":
                my_dict = {}
                for key, value in output_dict.items():
                    if key in ["sweep_params", "outputs"]:
                        my_dict[key] = list(value.keys())
            else:
                my_dict = output_dict

            with open(txt_fname, "w") as log_file:
                pprint.pprint(my_dict, log_file)

    def _write_output_to_h5(self, output_dict, h5_results_file_name):
        if self.config.h5_parent_group_name is None:
            # No parent groups exists, a new file will be created regardless
            f = h5py.File(h5_results_file_name, "w")
            parent_grp = f
        else:
            if os.path.isfile(h5_results_file_name):
                # File exists, we only need to add the new parent group
                f = h5py.File(h5_results_file_name, "a")
                parent_grp = f.require_group(self.config.h5_parent_group_name)
            else:
                # Create a new file since none exists and add the parent group
                f = h5py.File(h5_results_file_name, "w")
                parent_grp = f.require_group(self.config.h5_parent_group_name)

        for key, item in output_dict.items():
            grp = parent_grp.create_group(key)
            if key in ["sweep_params", "outputs"]:
                for subkey, subitem in item.items():
                    subgrp = grp.create_group(subkey)
                    for subsubkey, subsubitem in subitem.items():
                        if subsubkey[0] != "_":
                            if subsubkey == "lower bound" and subsubitem is None:
                                subgrp.create_dataset(subsubkey, data=np.finfo("d").min)
                            elif subsubkey == "upper bound" and subsubitem is None:
                                subgrp.create_dataset(subsubkey, data=np.finfo("d").max)
                            else:
                                subgrp.create_dataset(subsubkey, data=subsubitem)
            elif key in ["solve_successful", "nominal_idx", "differential_idx"]:
                grp.create_dataset(key, data=output_dict[key])

        f.close()

    def _write_to_csv(
        self,
        sweep_params,
        global_values,
        global_results_dict,
        global_results_arr,
        process_number,
    ):
        # Create the dataframe that is going to be written to a CSV
        global_save_data = np.hstack((global_values, global_results_arr))

        if process_number == self.parallel_manager.ROOT_PROCESS_RANK:
            data_header = ",".join(itertools.chain(sweep_params))
            for i, (key, item) in enumerate(global_results_dict["outputs"].items()):
                data_header = ",".join([data_header, key])

            if self.config["csv_results_file_name"] is not None:
                # Write the CSV
                np.savetxt(
                    self.config["csv_results_file_name"],
                    global_save_data,
                    header=data_header,
                    delimiter=",",
                    fmt="%.6e",
                )

                # If we want the interpolated output_list in CSV
                if self.config["interpolate_nan_outputs"]:
                    global_results_clean = self._interp_nan_values(
                        global_values, global_results_arr
                    )
                    global_save_data_clean = np.hstack(
                        (global_values, global_results_clean)
                    )

                    head, tail = os.path.split(self.config["csv_results_file_name"])

                    if head == "":
                        interp_file = "interpolated_%s" % (tail)
                    else:
                        interp_file = "%s/interpolated_%s" % (head, tail)

                    np.savetxt(
                        interp_file,
                        global_save_data_clean,
                        header=data_header,
                        delimiter=",",
                        fmt="%.6e",
                    )

        return global_save_data

    def save_results(
        self,
        sweep_params,
        local_values,
        global_values,
        local_results_dict,
        global_results_dict,
        global_results_arr,
        process_number,
    ):
        if process_number == self.parallel_manager.ROOT_PROCESS_RANK:
            if self.config["debugging_data_dir"] is not None:
                os.makedirs(self.config["debugging_data_dir"], exist_ok=True)
            if self.config["h5_results_file_name"] is not None:
                pathlib.Path(self.config["h5_results_file_name"]).parent.mkdir(
                    parents=True, exist_ok=True
                )
            if self.config["csv_results_file_name"] is not None:
                pathlib.Path(self.config["csv_results_file_name"]).parent.mkdir(
                    parents=True, exist_ok=True
                )

        self.parallel_manager.sync_with_peers()

        # Handle values in the debugging data_directory
        if self.config["debugging_data_dir"] is not None:
            self._write_debug_data(
                sweep_params,
                local_values,
                local_results_dict,
                self.config["h5_results_file_name"] is not None,
                self.config["csv_results_file_name"] is not None,
                process_number,
            )

        global_save_data = self._write_to_csv(
            sweep_params,
            global_values,
            global_results_dict,
            global_results_arr,
            process_number,
        )

        if (
            process_number == self.parallel_manager.ROOT_PROCESS_RANK
            and self.config["h5_results_file_name"] is not None
        ):
            # Save the data of output dictionary
            self._write_outputs(global_results_dict, txt_options="keys")

        return global_save_data



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import os
import idaes.logger as idaeslog
from pyomo.environ import (
    ConcreteModel,
    Expression,
    value,
    TransformationFactory,
    units as pyunits,
)
from pyomo.network import Arc, SequentialDecomposition
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.models.unit_models import Product
import idaes.core.util.scaling as iscale
from idaes.core import UnitModelCostingBlock

from watertap.core.util.initialization import assert_degrees_of_freedom, check_solve

from watertap.core.wt_database import Database
import watertap.core.zero_order_properties as prop_ZO
from watertap.unit_models.zero_order import (
    FeedZO,
    MABRZO,
    DMBRZO,
    PumpElectricityZO,
)
from watertap.costing.zero_order_costing import ZeroOrderCosting

# Set up logger
_log = idaeslog.getLogger(__name__)


def main():
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m, checkpoint="solve flowsheet after initializing system")
    display_reports(m)

    add_costing(m)
    initialize_costing(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    results = solve(m, checkpoint="solve flowsheet after costing")

    display_metrics_results(m)
    display_additional_results(m)

    return m, results


def build():
    # flowsheet set up
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.prop = prop_ZO.WaterParameterBlock(
        solute_list=["bod", "tss", "ammonium_as_nitrogen", "nitrate", "nitrogen"]
    )

    # unit models
    m.fs.feed = FeedZO(property_package=m.fs.prop)
    m.fs.mabr = MABRZO(property_package=m.fs.prop, database=m.db)
    m.fs.dmbr = DMBRZO(property_package=m.fs.prop, database=m.db)
    m.fs.P1 = PumpElectricityZO(property_package=m.fs.prop, database=m.db)

    m.fs.product_H2O = Product(property_package=m.fs.prop)

    # connections
    m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.P1.inlet)
    m.fs.s02 = Arc(source=m.fs.P1.outlet, destination=m.fs.mabr.inlet)
    m.fs.s03 = Arc(source=m.fs.mabr.treated, destination=m.fs.dmbr.inlet)
    m.fs.s04 = Arc(source=m.fs.dmbr.treated, destination=m.fs.product_H2O.inlet)
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scaling
    iscale.calculate_scaling_factors(m)

    return m


def set_operating_conditions(m):
    # ---specifications---
    # feed
    flow_vol = 417 * pyunits.m**3 / pyunits.hr
    conc_mass_bod = 103 * pyunits.mg / pyunits.liter
    conc_mass_tss = 100 * pyunits.mg / pyunits.liter
    conc_mass_ammonium_as_nitrogen = 26 * pyunits.mg / pyunits.liter
    conc_mass_nitrate = 1.3 * pyunits.mg / pyunits.liter

    m.fs.feed.flow_vol[0].fix(flow_vol)
    m.fs.feed.conc_mass_comp[0, "bod"].fix(conc_mass_bod)
    m.fs.feed.conc_mass_comp[0, "tss"].fix(conc_mass_tss)
    m.fs.feed.conc_mass_comp[0, "ammonium_as_nitrogen"].fix(
        conc_mass_ammonium_as_nitrogen
    )
    m.fs.feed.conc_mass_comp[0, "nitrate"].fix(conc_mass_nitrate)
    m.fs.feed.conc_mass_comp[0, "nitrogen"].fix(1e-5)
    solve(m.fs.feed, checkpoint="solve feed block")

    # pump
    m.fs.P1.load_parameters_from_database(use_default_removal=True)
    m.fs.P1.lift_height.fix(1)

    # mabr
    m.fs.mabr.load_parameters_from_database(use_default_removal=True)

    # dmbr
    m.fs.dmbr.load_parameters_from_database(use_default_removal=True)


def initialize_system(m):
    seq = SequentialDecomposition()
    seq.options.tear_set = []
    seq.options.iterLim = 1
    seq.run(m, lambda u: u.initialize())


def solve(blk, solver=None, checkpoint=None, tee=False, fail_flag=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    check_solve(results, checkpoint=checkpoint, logger=_log, fail_flag=fail_flag)
    return results


def display_reports(m):
    unit_list = ["feed", "P1", "mabr", "dmbr"]
    for u in unit_list:
        m.fs.component(u).report()


def add_costing(m):
    source_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "biomembrane_filtration_global_costing.yaml",
    )
    m.fs.costing = ZeroOrderCosting(case_study_definition=source_file)
    # typing aid
    costing_kwargs = {"flowsheet_costing_block": m.fs.costing}
    m.fs.mabr.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.dmbr.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.P1.costing = UnitModelCostingBlock(**costing_kwargs)

    m.fs.costing.cost_process()
    m.fs.costing.add_electricity_intensity(m.fs.product_H2O.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.product_H2O.properties[0].flow_vol)

    # other levelized costs
    m.fs.costing.annual_water_inlet = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.feed.properties[0].flow_vol,
            to_units=pyunits.m**3 / m.fs.costing.base_period,
        )
    )

    m.fs.costing.annual_water_production = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.product_H2O.properties[0].flow_vol,
            to_units=pyunits.m**3 / m.fs.costing.base_period,
        )
    )

    m.fs.costing.total_annualized_cost = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
    )

    m.fs.costing.LCOT = Expression(
        expr=(m.fs.costing.total_annualized_cost / m.fs.costing.annual_water_inlet),
        doc="Levelized Cost of Treatment",
    )


def initialize_costing(m):
    m.fs.costing.initialize()


def display_metrics_results(m):
    print("----------Levelized costs----------")
    LCOT = value(
        pyunits.convert(
            m.fs.costing.LCOT, to_units=m.fs.costing.base_currency / pyunits.m**3
        )
    )
    print(f"Levelized Cost of Treatment: {LCOT:.4f} $/m3 of feed")
    LCOW = value(
        pyunits.convert(
            m.fs.costing.LCOW, to_units=m.fs.costing.base_currency / pyunits.m**3
        )
    )
    print(f"Levelized Cost of Water: {LCOW:.4f} $/m3 of product")

    print("----------Capital costs----------")
    DCC_normalized = value(
        pyunits.convert(
            (
                m.fs.mabr.costing.direct_capital_cost
                + m.fs.dmbr.costing.direct_capital_cost
                + m.fs.P1.costing.direct_capital_cost
            )
            / m.fs.feed.properties[0].flow_vol,
            to_units=m.fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        )
    )
    print(f"Normalized direct capital costs: {DCC_normalized:.4f} $/(m3/day)")
    ICC_normalized = value(
        pyunits.convert(
            m.fs.costing.total_capital_cost / m.fs.feed.properties[0].flow_vol,
            to_units=m.fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        )
    )
    print(f"Normalized total capital costs: {ICC_normalized:.4f} $/(m3/day)")

    print("----------Operating costs----------")
    FMC_normalized = value(
        pyunits.convert(
            m.fs.costing.maintenance_cost / m.fs.costing.total_capital_cost,
            to_units=1 / pyunits.a,
        )
    )
    print(f"Normalized maintenance costs: {FMC_normalized:.4f} 1/year")
    BRC_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_fixed_operating_cost
            / m.fs.costing.total_capital_cost,
            to_units=1 / pyunits.a,
        )
    )
    print(f"Normalized bead replacement cost: {BRC_normalized:.4f} 1/year")
    EC_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["electricity"]
            / m.fs.costing.annual_water_inlet,
            to_units=m.fs.costing.base_currency / pyunits.m**3,
        )
    )
    print(f"Normalized electricity cost: {EC_normalized:.4f} $/m3 of feed")

    print("----------Performance metrics----------")
    volumetric_recovery = value(
        m.fs.product_H2O.properties[0].flow_vol / m.fs.feed.properties[0].flow_vol
    )
    print(f"Water recovery: {volumetric_recovery:.4f} m3 of product/m3 of feed")
    BODR_normalized = value(
        pyunits.convert(
            1
            - m.fs.product_H2O.properties[0].flow_mass_comp["bod"]
            / m.fs.feed.properties[0].flow_mass_comp["bod"],
            to_units=pyunits.dimensionless,
        )
    )
    print(f"BOD removal: {BODR_normalized:.4f} dimensionless")
    TSSR_normalized = value(
        pyunits.convert(
            1
            - m.fs.product_H2O.properties[0].flow_mass_comp["tss"]
            / m.fs.feed.properties[0].flow_mass_comp["tss"],
            to_units=pyunits.dimensionless,
        )
    )
    print(f"TSS removal: {TSSR_normalized:.4f} dimensionless")
    NR_normalized = value(
        pyunits.convert(
            1
            - (
                m.fs.product_H2O.properties[0].flow_mass_comp["ammonium_as_nitrogen"]
                + m.fs.product_H2O.properties[0].flow_mass_comp["nitrate"]
            )
            / (
                m.fs.feed.properties[0].flow_mass_comp["ammonium_as_nitrogen"]
                + m.fs.feed.properties[0].flow_mass_comp["nitrate"]
            ),
            to_units=pyunits.dimensionless,
        )
    )
    print(f"Total N removal: {NR_normalized:.4f} dimensionless")

    print("----------Energy intensity----------")
    SEC = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_electricity / m.fs.feed.properties[0].flow_vol,
            to_units=pyunits.kWh / pyunits.m**3,
        )
    )
    print(f"Specific electricity consumption: {SEC:.4f} kWh/m3 of feed")


def display_additional_results(m):
    print("----------Outlets----------")
    product_H2O_flow = value(
        pyunits.convert(
            m.fs.product_H2O.properties[0].flow_vol,
            to_units=pyunits.m**3 / pyunits.hr,
        )
    )
    print(f"H2O outlet flow: {product_H2O_flow:.4f} m3/h")
    product_H2O_BOD = value(
        pyunits.convert(
            m.fs.product_H2O.properties[0].conc_mass_comp["bod"],
            to_units=pyunits.g / pyunits.L,
        )
    )
    print(f"H2O outlet BOD conc: {product_H2O_BOD:.4f} g/L")
    product_H2O_TSS = value(
        pyunits.convert(
            m.fs.product_H2O.properties[0].conc_mass_comp["tss"],
            to_units=pyunits.g / pyunits.L,
        )
    )
    print(f"H2O outlet TSS conc: {product_H2O_TSS:.4f} g/L")
    product_H2O_total_nitrogen = value(
        pyunits.convert(
            (
                m.fs.product_H2O.properties[0].conc_mass_comp["ammonium_as_nitrogen"]
                + m.fs.product_H2O.properties[0].conc_mass_comp["nitrate"]
            ),
            to_units=pyunits.g / pyunits.L,
        )
    )
    print(f"H2O outlet total nitrogen conc: {product_H2O_total_nitrogen:.4f} g/L")

    print("----------Capital costs----------")
    total_capital_costs = value(m.fs.costing.total_capital_cost) / 1e6
    print(f"Total capital costs: {total_capital_costs:.4f} $M")
    mabr_capital_costs = value(m.fs.mabr.costing.capital_cost) / 1e6
    print(f"MABR capital costs: {mabr_capital_costs:.4f} $M")
    dmbr_capital_costs = value(m.fs.dmbr.costing.capital_cost) / 1e6
    print(f"DMBR capital costs: {dmbr_capital_costs:.4f} $M")
    pump_capital_costs = value(m.fs.P1.costing.capital_cost) / 1e6
    print(f"Pump capital costs: {pump_capital_costs:.4f} $M")

    print("----------Operating costs----------")
    total_operating_costs = value(m.fs.costing.total_operating_cost) / 1e6
    print(f"Total operating costs: {total_operating_costs:.4f} $M/year")
    fixed_operating_costs = value(m.fs.costing.total_fixed_operating_cost) / 1e6
    print(f"Fixed operating costs: {fixed_operating_costs:.4f} $M/year")
    electricity_operating_costs = (
        value(m.fs.costing.aggregate_flow_costs["electricity"]) / 1e3
    )
    print(f"Electricity operating costs: {electricity_operating_costs:.4f} $k/year")


if __name__ == "__main__":
    m, results = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""
Tests for zero-order reactive anaerobic digestion model.
"""

import pytest

from pyomo.environ import (
    Block,
    Var,
    Constraint,
    ConcreteModel,
    value,
    assert_optimal_termination,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import AnaerobicDigestionReactiveZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestAnaerobicDigestionReactiveZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=["tss", "cod", "tkn", "acetic_acid", "ammonium_as_nitrogen"]
        )

        m.fs.unit = AnaerobicDigestionReactiveZO(
            property_package=m.fs.params, database=m.db
        )

        # Inlet mass flowrates in kg/s
        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(100)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(0.4)
        m.fs.unit.inlet.flow_mass_comp[0, "cod"].fix(0.35)
        m.fs.unit.inlet.flow_mass_comp[0, "tkn"].fix(0.01)
        m.fs.unit.inlet.flow_mass_comp[0, "acetic_acid"].fix(0)
        m.fs.unit.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(0)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db
        assert model.fs.unit._tech_type == "anaerobic_digestion_reactive"
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.biogas_tss_ratio, Var)
        assert isinstance(model.fs.unit.biogas_production, Var)
        assert isinstance(model.fs.unit.biogas_prod, Constraint)

    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("anaerobic_digestion_reactive")
        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"].keys():
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.biogas_tss_ratio.fixed
        assert model.fs.unit.biogas_tss_ratio.value == data["biogas_tss_ratio"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(0.16, rel=1e-3) == value(
            model.fs.unit.biogas_production[0]
        )

        assert pytest.approx(0.10076, rel=1e-3) == value(
            model.fs.unit.properties_in[0].flow_vol
        )
        assert pytest.approx(3.9698, rel=1e-3) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(3.4736, rel=1e-3) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["cod"]
        )
        assert pytest.approx(0.099246, rel=1e-3) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["tkn"]
        )
        assert pytest.approx(0, rel=1e-3) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["acetic_acid"]
        )
        assert pytest.approx(0, rel=1e-3) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["ammonium_as_nitrogen"]
        )

        assert pytest.approx(0.1, rel=1e-3) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert value(model.fs.unit.properties_treated[0].conc_mass_comp["tss"]) < 1e-6
        assert value(model.fs.unit.properties_treated[0].conc_mass_comp["cod"]) < 1e-6
        assert value(model.fs.unit.properties_treated[0].conc_mass_comp["tkn"]) < 1e-6
        assert pytest.approx(0.199856, rel=1e-3) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["acetic_acid"]
        )
        assert pytest.approx(0.519626, rel=1e-3) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["ammonium_as_nitrogen"]
        )

        assert pytest.approx(0.00038, rel=1e-3) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(0.2, rel=1e-3) == value(
            model.fs.unit.properties_byproduct[0].flow_mass_comp["tss"]
        )
        assert pytest.approx(0.175, rel=1e-3) == value(
            model.fs.unit.properties_byproduct[0].flow_mass_comp["cod"]
        )
        assert pytest.approx(0.005, rel=1e-3) == value(
            model.fs.unit.properties_byproduct[0].flow_mass_comp["tkn"]
        )
        assert (
            value(model.fs.unit.properties_byproduct[0].conc_mass_comp["acetic_acid"])
            < 1e-6
        )
        assert (
            value(
                model.fs.unit.properties_byproduct[0].conc_mass_comp[
                    "ammonium_as_nitrogen"
                ]
            )
            < 1e-6
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    + sum(
                        model.fs.unit.generation_rxn_comp[0, r, j]
                        for r in model.fs.unit.reaction_set
                    )
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):
        model.fs.unit.report()


def test_costing():
    m = ConcreteModel()
    m.db = Database()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = WaterParameterBlock(
        solute_list=["tss", "cod", "tkn", "acetic_acid", "ammonium_as_nitrogen"]
    )
    m.fs.costing = ZeroOrderCosting()
    m.fs.unit = AnaerobicDigestionReactiveZO(
        property_package=m.fs.params, database=m.db
    )

    # Inlet mass flowrates in kg/s
    m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(100)
    m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(0.4)
    m.fs.unit.inlet.flow_mass_comp[0, "cod"].fix(0.35)
    m.fs.unit.inlet.flow_mass_comp[0, "tkn"].fix(0.01)
    m.fs.unit.inlet.flow_mass_comp[0, "acetic_acid"].fix(0)
    m.fs.unit.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(0)

    m.db.get_unit_operation_parameters("anaerobic_digestion_reactive")
    m.fs.unit.load_parameters_from_database(use_default_removal=True)

    assert degrees_of_freedom(m.fs.unit) == 0

    m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.costing.cost_process()

    assert isinstance(m.fs.costing.anaerobic_digestion_reactive, Block)
    assert isinstance(
        m.fs.costing.anaerobic_digestion_reactive.capital_a_parameter, Var
    )
    assert isinstance(
        m.fs.costing.anaerobic_digestion_reactive.capital_b_parameter, Var
    )
    assert isinstance(m.fs.costing.anaerobic_digestion_reactive.reference_state, Var)
    assert isinstance(m.fs.unit.costing.capital_cost, Var)
    assert isinstance(m.fs.unit.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit) == 0
    initialization_tester(m)

    assert pytest.approx(18.9601, rel=1e-3) == value(m.fs.unit.costing.capital_cost)

    assert m.fs.unit.electricity[0] in m.fs.costing._registered_flows["electricity"]
    assert isinstance(m.fs.costing.total_capital_cost, Var)
    assert isinstance(m.fs.costing.aggregate_flow_costs, Var)


db = Database()
params = db._get_technology("anaerobic_digestion_reactive")


class TestAnaerobicDigestionReactivesubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["tss", "cod", "tkn"])

        m.fs.unit = AnaerobicDigestionReactiveZO(
            property_package=m.fs.params, database=db
        )

        return m

    @pytest.mark.parametrize("subtype", [k for k in params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters(
            "anaerobic_digestion_reactive", subtype=subtype
        )

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            if j not in data["removal_frac_mass_comp"].keys():
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]


@pytest.mark.component
def test_costing_GLSD():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(
        solute_list=["tss", "methane", "carbon_dioxide", "nitrogen", "oxygen"]
    )

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = AnaerobicDigestionReactiveZO(
        property_package=m.fs.params,
        database=m.db,
        process_subtype="GLSD_anaerobic_digester",
    )

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(10)
    m.fs.unit1.inlet.flow_mass_comp[0, "methane"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "carbon_dioxide"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "nitrogen"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "oxygen"].fix(1)

    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.anaerobic_digestion_reactive, Block)
    assert isinstance(
        m.fs.costing.anaerobic_digestion_reactive.capital_a_parameter, Var
    )
    assert isinstance(
        m.fs.costing.anaerobic_digestion_reactive.capital_b_parameter, Var
    )
    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)
    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from idaes.core.solvers import get_solver
from pyomo.environ import value, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.amo_1575_magprex.magprex import (
    build,
    set_operating_conditions,
    initialize_system,
    solve,
    add_costing,
    display_results,
    display_costing,
)

solver = get_solver()


class TestMagprexFlowsheet:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build()
        return m  # return model

    @pytest.mark.unit()
    def test_build(self, system_frame):
        m = system_frame
        assert_units_consistent(m)
        assert_degrees_of_freedom(m, 14)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        initialize_system(m)

        # test feed water flow
        assert value(m.fs.feed.properties[0].flow_mass_comp["H2O"]) == pytest.approx(
            35.043286, rel=1e-3
        )
        assert value(
            m.fs.feed.properties[0].flow_mass_comp["phosphates"]
        ) == pytest.approx(0.0122694, rel=1e-5)
        assert value(
            m.fs.feed.properties[0].flow_mass_comp["struvite"]
        ) == pytest.approx(0, abs=1e-10)

    @pytest.mark.component
    def test_solve(self, system_frame):
        m = system_frame
        results = solve(m)
        assert_optimal_termination(results)

        # check two struvite product flow
        assert value(
            m.fs.struvite_product.properties[0].flow_mass_comp["H2O"]
        ) == pytest.approx(4.302064e-21, rel=1e-3)
        assert value(
            m.fs.struvite_product.properties[0].flow_mass_comp["struvite"]
        ) == pytest.approx(0.01165597, rel=1e-3)
        assert value(
            m.fs.struvite_product.properties[0].flow_mass_comp["phosphates"]
        ) == pytest.approx(0, abs=1e-6)

        # check two biosolid product flow
        assert value(
            m.fs.biosolid_product.properties[0].flow_mass_comp["H2O"]
        ) == pytest.approx(4.302064e-21, rel=1e-3)
        assert value(
            m.fs.biosolid_product.properties[0].flow_mass_comp["struvite"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.biosolid_product.properties[0].flow_mass_comp["phosphates"]
        ) == pytest.approx(0.000582799, rel=1e-3)

        # check two centrate flow
        assert value(
            m.fs.centrate.properties[0].flow_mass_comp["H2O"]
        ) == pytest.approx(35.0432861, rel=1e-3)
        assert value(
            m.fs.centrate.properties[0].flow_mass_comp["struvite"]
        ) == pytest.approx(0, abs=1e-6)
        assert value(
            m.fs.centrate.properties[0].flow_mass_comp["phosphates"]
        ) == pytest.approx(3.067361e-5, rel=1e-3)

    @pytest.mark.component
    def test_costing(self, system_frame):
        m = system_frame

        add_costing(m)
        m.fs.costing.initialize()
        results = solve(m)

        assert_optimal_termination(results)

        # check costing
        assert value(m.fs.costing.LCOW) == pytest.approx(
            2.170813986, rel=1e-3
        )  # in $/m**3
        assert value(m.fs.costing.LCOS) == pytest.approx(6.528763868, rel=1e-3)

    @pytest.mark.component
    def test_display(self, system_frame):
        m = system_frame
        display_results(m)
        display_costing(m)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains utility functions for infeasibility diagnostics of WaterTAP models.
"""
import logging, sys

from contextlib import contextmanager
from math import isclose
from pyomo.environ import Var, Constraint, value
from pyomo.util.infeasible import log_infeasible_constraints, log_infeasible_bounds

_logger = logging.getLogger("watertap.core.util.infeasible.print")
_logger.setLevel(logging.DEBUG)


@contextmanager
def _logging_handler(output_file):
    """
    helper for removing the complexity of dealing with loggers for the
    utility functions in this file
    """
    if output_file is None:
        _handler = logging.StreamHandler(stream=sys.stdout)
    else:
        _handler = logging.FileHandler(output_file, mode="w")
    _handler.setFormatter(logging.Formatter("%(message)s"))
    _logger.addHandler(_handler)
    try:
        yield _logger
    finally:
        _logger.removeHandler(_handler)


def print_infeasible_constraints(
    m, tol=1e-6, print_expression=False, print_variables=False, output_file=None
):
    """
    print the infeasble constraints in the model

    Args:
        m : A Pyomo Block or ConcreteModel
        tol : (optional) absolute feasibility tolerance, default 1e-06
        print_expressions : (optional) If True, prints the constraint expression
                            (default: False)
        print_variables : (optional) If True, prints the constraint variable names
                          and values (default: False)
        output_file : (optional) file to write results to. If None (default)
                      print infeasible variable bounds to the screen.

    Returns:
        None
    """
    with _logging_handler(output_file) as logger:
        log_infeasible_constraints(
            m,
            tol=tol,
            logger=logger,
            log_expression=print_expression,
            log_variables=print_variables,
        )


def print_infeasible_bounds(m, tol=1e-6, output_file=None):
    """
    print the infeasible variable bounds in the model

    Args:
        m : A Pyomo Block or ConcreteModel
        tol : (optional) absolute feasibility tolerance, default 1e-06
        output_file : (optional) file to write results to. If None (default)
                      print infeasible variable bounds to the screen.

    Returns:
        None
    """
    with _logging_handler(output_file) as logger:
        log_infeasible_bounds(m, tol=tol, logger=logger)


def print_variables_close_to_bounds(m, rel_tol=1e-4, abs_tol=1e-12):
    """
    print variables close to their bounds

    Args:
        m : A Pyomo Block or ConcreteModel
        rel_tol : (optional) relative tolerance for comparing the value
                   to the bound, default 1e-04
        abs_tol : (optional) absolute tolerance for comparing the value
                   to the bound, default 1e-12

    Returns:
        None
    """
    for var in m.component_data_objects(ctype=Var, descend_into=True):
        if var.fixed:
            continue
        val = var.value
        if val is None:
            print(f"No value for Var {var.name}")
        else:
            _eval_close(var, val, rel_tol, abs_tol)


def print_constraints_close_to_bounds(m, rel_tol=1e-4, abs_tol=1e-5):
    """
    print constraints close to their bounds

    Args:
        m : A Pyomo Block or ConcreteModel
        rel_tol : (optional) relative tolerance for comparing the value
                   to the bound, default 1e-04
        abs_tol : (optional) absolute tolerance for comparing the value
                   to the bound, default 1e-05

    Returns:
        None
    """
    for con in m.component_data_objects(
        ctype=Constraint, descend_into=True, active=True
    ):
        if con.equality:
            continue
        val = value(con.body, exception=False)
        if val is None:
            print(f"Cannot evaluate Constraint {con.name}: missing variable value")
        else:
            _eval_close(con, val, rel_tol, abs_tol)


def print_close_to_bounds(m, rel_tol=1e-04, abs_tol=1e-12):
    """
    Print variables and constraints which are near their bounds

    Args:
        m : A Pyomo Block or ConcreteModel
        rel_tol : (optional) relative tolerance for comparing the value
                   to the bound, default 1e-04
        abs_tol : (optional) absolute tolerance for comparing the value
                   to the bound, default 1e-12

    Returns:
        None
    """
    print_variables_close_to_bounds(m, rel_tol=rel_tol, abs_tol=abs_tol)
    print_constraints_close_to_bounds(m, rel_tol=rel_tol, abs_tol=abs_tol)


def _eval_close(obj, val, rel_tol, abs_tol):
    """
    Helper for evaluating bounds and printing Pyomo variable or constraint.
    Prints if the Pyomo object is close to one or both of its bounds.

    Args:
        obj : Pyomo object (_VarData or _ConstraintData)
        val : the reference value
        rel_tol : the relative tolerance for math.isclose
        abs_tol : the absolute tolerance for math.isclose

    Returns:
        None
    """
    lb = value(obj.lb)
    ub = value(obj.ub)
    if (
        lb is not None
        and ub is not None
        and isclose(lb, ub, rel_tol=rel_tol, abs_tol=abs_tol)
    ):
        return
    if lb is not None and isclose(lb, val, rel_tol=rel_tol, abs_tol=abs_tol):
        print(f"{obj.name} near LB of {lb}")
    if ub is not None and isclose(val, ub, rel_tol=rel_tol, abs_tol=abs_tol):
        print(f"{obj.name} near UB of {ub}")



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import os
import numpy as np
import pyomo.environ as pyo

from pyomo.environ import value

from watertap.tools.parameter_sweep.sampling_types import *
from watertap.tools.parameter_sweep import ParameterSweep, parameter_sweep
from watertap.tools.parameter_sweep.parameter_sweep_writer import *

import watertap.tools.MPI as MPI
from watertap.tools.parallel.parallel_manager_factory import has_mpi_peer_processes

# -----------------------------------------------------------------------------


def build_model():
    m = pyo.ConcreteModel()
    m.fs = fs = pyo.Block()

    fs.input = pyo.Var(["a", "b"], within=pyo.UnitInterval, initialize=0.5)
    fs.output = pyo.Var(["c", "d"], within=pyo.UnitInterval, initialize=0.5)

    fs.slack = pyo.Var(["ab_slack", "cd_slack"], bounds=(0, 0), initialize=0.0)
    fs.slack_penalty = pyo.Param(default=1000.0, mutable=True, within=pyo.PositiveReals)

    fs.ab_constr = pyo.Constraint(
        expr=(fs.output["c"] + fs.slack["ab_slack"] == 2 * fs.input["a"])
    )
    fs.cd_constr = pyo.Constraint(
        expr=(fs.output["d"] + fs.slack["cd_slack"] == 3 * fs.input["b"])
    )

    fs.performance = pyo.Expression(expr=pyo.summation(fs.output))

    m.objective = pyo.Objective(
        expr=m.fs.performance - m.fs.slack_penalty * pyo.summation(m.fs.slack),
        sense=pyo.maximize,
    )
    return m


def build_model_for_tps():
    model = build_model()
    model.fs.slack_penalty = 1000.0
    model.fs.slack.setub(0)
    return model


def build_sweep_params_for_tps(model):
    A = model.fs.input["a"]
    B = model.fs.input["b"]
    sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}
    return sweep_params


def build_outputs_for_tps(model, output_keys):
    outputs = {}
    for key, pyo_object in output_keys.items():
        outputs[key] = model.find_component(pyo_object)
    return outputs


def dummy_kernel_logic(solution_succesful):
    init_state = [True]
    solved_state = [False]
    for sf in solution_succesful:
        if sf and init_state[-1]:
            # we solved model from init and/or prior solved state
            init_state.append(True)
            solved_state.append(True)
        elif sf and init_state[-1] == False:
            # we try to solve ,but first init
            init_state.append(True)
            solved_state.append(False)
            # solve is succesful
            init_state.append(True)
            solved_state.append(True)
        else:
            if solved_state[-1]:
                # this means solution failed after
                # on model that was solved with applied sweep params
                # we add failed states
                init_state.append(False)
                solved_state.append(False)
                # kernel reinits model and then tries solving again
                init_state.append(True)
                solved_state.append(False)
                # but it fails again
                init_state.append(False)
                solved_state.append(False)
            else:
                # we are solving from failed solve, so we reinit
                init_state.append(True)
                solved_state.append(False)
                # but fail again
                init_state.append(False)
                solved_state.append(False)
                # kernel solving from inited state, so we failed we move on
    return init_state, solved_state


class TestParameterSweep:
    @pytest.fixture(scope="class")
    def model(self):
        return build_model()

    @pytest.mark.unit
    def test_single_index_unrolled(self):
        indexed_var = pyo.Var(["a"])
        indexed_var.construct()

        ls = LinearSample(indexed_var, None, None, None)

        assert ls.pyomo_object is indexed_var["a"]

    @pytest.mark.unit
    def test_multiple_indices_error(self):
        indexed_var = pyo.Var(["a", "b"])
        indexed_var.construct()

        with pytest.raises(Exception):
            ls = LinearSample(indexed_var, None, None, None)

    @pytest.mark.component
    def test_linear_build_combinations(self):
        ps = ParameterSweep()

        A_param = pyo.Param(initialize=0.0, mutable=True)
        B_param = pyo.Param(initialize=1.0, mutable=True)
        C_param = pyo.Param(initialize=2.0, mutable=True)

        range_A = [0.0, 10.0]
        range_B = [1.0, 20.0]
        range_C = [2.0, 30.0]

        nn_A = 4
        nn_B = 5
        nn_C = 6

        param_dict = dict()
        param_dict["var_A"] = LinearSample(A_param, range_A[0], range_A[1], nn_A)
        param_dict["var_B"] = LinearSample(B_param, range_B[0], range_B[1], nn_B)
        param_dict["var_C"] = LinearSample(C_param, range_C[0], range_C[1], nn_C)

        global_combo_array = ps._build_combinations(
            param_dict, SamplingType.FIXED, None
        )

        assert np.shape(global_combo_array)[0] == nn_A * nn_B * nn_C
        assert np.shape(global_combo_array)[1] == len(param_dict)

        assert global_combo_array[0, 0] == pytest.approx(range_A[0])
        assert global_combo_array[0, 1] == pytest.approx(range_B[0])
        assert global_combo_array[0, 2] == pytest.approx(range_C[0])

        assert global_combo_array[-1, 0] == pytest.approx(range_A[1])
        assert global_combo_array[-1, 1] == pytest.approx(range_B[1])
        assert global_combo_array[-1, 2] == pytest.approx(range_C[1])

    @pytest.mark.component
    def test_geom_build_combinations(self):
        ps = ParameterSweep()

        A_param = pyo.Param(initialize=0.0, mutable=True)
        B_param = pyo.Param(initialize=1.0, mutable=True)
        C_param = pyo.Param(initialize=2.0, mutable=True)

        range_A = [1.0, 10.0]
        range_B = [2.0, 20.0]
        range_C = [3.0, 30.0]

        nn_A = 2
        nn_B = 3
        nn_C = 4

        param_dict = dict()
        param_dict["var_A"] = GeomSample(A_param, range_A[0], range_A[1], nn_A)
        param_dict["var_B"] = GeomSample(B_param, range_B[0], range_B[1], nn_B)
        param_dict["var_C"] = GeomSample(C_param, range_C[0], range_C[1], nn_C)

        global_combo_array = ps._build_combinations(
            param_dict, SamplingType.FIXED, None
        )

        assert np.shape(global_combo_array)[0] == nn_A * nn_B * nn_C
        assert np.shape(global_combo_array)[1] == len(param_dict)

        assert global_combo_array[0, 0] == pytest.approx(range_A[0])
        assert global_combo_array[0, 1] == pytest.approx(range_B[0])
        assert global_combo_array[0, 2] == pytest.approx(range_C[0])

        assert global_combo_array[-1, 0] == pytest.approx(range_A[1])
        assert global_combo_array[-1, 1] == pytest.approx(range_B[1])
        assert global_combo_array[-1, 2] == pytest.approx(range_C[1])

    @pytest.mark.component
    def test_reverse_geom_build_combinations(self):
        ps = ParameterSweep()

        A_param = pyo.Param(initialize=0.0, mutable=True)
        B_param = pyo.Param(initialize=1.0, mutable=True)
        C_param = pyo.Param(initialize=2.0, mutable=True)

        range_A = [1.0, 10.0]
        range_B = [2.0, 20.0]
        range_C = [3.0, 30.0]

        nn_A = 2
        nn_B = 3
        nn_C = 4

        param_dict = dict()
        param_dict["var_A"] = ReverseGeomSample(A_param, range_A[0], range_A[1], nn_A)
        param_dict["var_B"] = ReverseGeomSample(B_param, range_B[0], range_B[1], nn_B)
        param_dict["var_C"] = ReverseGeomSample(C_param, range_C[0], range_C[1], nn_C)

        global_combo_array = ps._build_combinations(
            param_dict, SamplingType.FIXED, None
        )

        assert np.shape(global_combo_array)[0] == nn_A * nn_B * nn_C
        assert np.shape(global_combo_array)[1] == len(param_dict)

        assert global_combo_array[0, 0] == pytest.approx(range_A[0])
        assert global_combo_array[0, 1] == pytest.approx(range_B[0])
        assert global_combo_array[0, 2] == pytest.approx(range_C[0])

        assert global_combo_array[-1, 0] == pytest.approx(range_A[1])
        assert global_combo_array[-1, 1] == pytest.approx(range_B[1])
        assert global_combo_array[-1, 2] == pytest.approx(range_C[1])

    @pytest.mark.component
    def test_predetermined_fixed_build_combinations(self):
        ps = ParameterSweep()

        A_param = pyo.Param(initialize=0.0, mutable=True)
        B_param = pyo.Param(initialize=1.0, mutable=True)
        C_param = pyo.Param(initialize=2.0, mutable=True)

        nn_A = 2
        nn_B = 3
        nn_C = 4

        A_values = nn_A + np.arange(nn_A)
        B_values = nn_B + np.arange(nn_B)
        C_values = nn_C + np.arange(nn_C)

        param_dict = dict()
        param_dict["var_A"] = PredeterminedFixedSample(A_param, A_values)
        param_dict["var_B"] = PredeterminedFixedSample(B_param, B_values)
        param_dict["var_C"] = PredeterminedFixedSample(C_param, C_values)

        global_combo_array = ps._build_combinations(
            param_dict, SamplingType.FIXED, None
        )

        assert np.shape(global_combo_array)[0] == nn_A * nn_B * nn_C
        assert np.shape(global_combo_array)[1] == len(param_dict)
        assert (global_combo_array[0] == np.array([nn_A, nn_B, nn_C])).all()
        assert (global_combo_array[-1] == 2 * np.array([nn_A, nn_B, nn_C]) - 1).all()

    @pytest.mark.component
    def test_status_publishing(self):
        requests = pytest.importorskip(
            "requests",
            reason="requests (parameter_sweep optional dependency) not available",
        )
        ps = ParameterSweep(
            publish_progress=True, publish_address="http://localhost:8888"
        )

        with pytest.raises(requests.exceptions.ConnectionError):
            r = ps._publish_updates(1, True, 5.0)

    def test_random_build_combinations(self):
        ps = ParameterSweep()

        nn = int(1e5)

        # Uniform random sampling [lower_limit, upper_limit]
        A_param = pyo.Param(initialize=-10.0, mutable=True)
        B_param = pyo.Param(initialize=0.0, mutable=True)
        C_param = pyo.Param(initialize=10.0, mutable=True)

        range_A = [-10.0, 0.0]
        range_B = [0.0, 10.0]
        range_C = [10.0, 20.0]

        param_dict = dict()
        param_dict["var_A"] = UniformSample(A_param, range_A[0], range_A[1], nn)
        param_dict["var_B"] = UniformSample(B_param, range_B[0], range_B[1], nn)
        param_dict["var_C"] = UniformSample(C_param, range_C[0], range_C[1], nn)

        global_combo_array = ps._build_combinations(param_dict, SamplingType.RANDOM, nn)

        assert np.shape(global_combo_array)[0] == nn
        assert np.shape(global_combo_array)[1] == len(param_dict)

        assert np.all(range_A[0] < global_combo_array[:, 0])
        assert np.all(range_B[0] < global_combo_array[:, 1])
        assert np.all(range_C[0] < global_combo_array[:, 2])

        assert np.all(global_combo_array[:, 0] < range_A[1])
        assert np.all(global_combo_array[:, 1] < range_B[1])
        assert np.all(global_combo_array[:, 2] < range_C[1])

        # Normal random sampling [mean, stdev]
        A_param = pyo.Param(initialize=10.0, mutable=True)
        B_param = pyo.Param(initialize=100.0, mutable=True)
        C_param = pyo.Param(initialize=1000.0, mutable=True)

        range_A = [10.0, 5.0]
        range_B = [100.0, 50.0]
        range_C = [1000.0, 0.0]

        param_dict = dict()
        param_dict["var_A"] = NormalSample(A_param, range_A[0], range_A[1], nn)
        param_dict["var_B"] = NormalSample(B_param, range_B[0], range_B[1], nn)
        param_dict["var_C"] = NormalSample(C_param, range_C[0], range_C[1], nn)

        global_combo_array = ps._build_combinations(param_dict, SamplingType.RANDOM, nn)

        assert np.shape(global_combo_array)[0] == nn
        assert np.shape(global_combo_array)[1] == len(param_dict)

        assert np.mean(global_combo_array[:, 0]) < (range_A[0] + range_A[1])
        assert np.mean(global_combo_array[:, 1]) < (range_B[0] + range_B[1])

        assert (range_A[0] - range_A[1]) < np.mean(global_combo_array[:, 0])
        assert (range_B[0] - range_B[1]) < np.mean(global_combo_array[:, 1])

        assert np.all(global_combo_array[:, 2] == range_C[0])

        A_param = pyo.Param(initialize=0.0, mutable=True)
        B_param = pyo.Param(initialize=0.0, mutable=True)
        C_param = pyo.Param(initialize=0.0, mutable=True)

        A_values = np.arange(nn)
        B_values = 10.0 + A_values
        C_values = 20.0 + A_values

        param_dict = dict()
        param_dict["var_A"] = PredeterminedRandomSample(A_param, A_values)
        param_dict["var_B"] = PredeterminedRandomSample(B_param, B_values)
        param_dict["var_C"] = PredeterminedRandomSample(C_param, C_values)

        global_combo_array = ps._build_combinations(param_dict, SamplingType.RANDOM, nn)

        assert np.shape(global_combo_array)[0] == nn
        assert (global_combo_array == np.array([A_values, B_values, C_values]).T).all()

    @pytest.mark.component
    def test_divide_combinations(self):
        ps = ParameterSweep()

        A_param = pyo.Param(initialize=0.0, mutable=True)
        B_param = pyo.Param(initialize=1.0, mutable=True)
        C_param = pyo.Param(initialize=2.0, mutable=True)

        range_A = [0.0, 10.0]
        range_B = [1.0, 20.0]
        range_C = [2.0, 30.0]

        nn_A = 4
        nn_B = 5
        nn_C = 6

        param_dict = dict()
        param_dict["var_A"] = LinearSample(A_param, range_A[0], range_A[1], nn_A)
        param_dict["var_B"] = LinearSample(B_param, range_B[0], range_B[1], nn_B)
        param_dict["var_C"] = LinearSample(C_param, range_C[0], range_C[1], nn_C)

        global_combo_array = ps._build_combinations(
            param_dict, SamplingType.FIXED, None
        )

        num_procs = ps.parallel_manager.number_of_worker_processes()
        rank = ps.parallel_manager.get_rank()
        test = np.array_split(global_combo_array, num_procs, axis=0)[rank]

        local_combo_array = ps._divide_combinations(global_combo_array)

        assert np.shape(local_combo_array)[1] == 3

        assert np.allclose(test[:, 0], local_combo_array[:, 0])
        assert np.allclose(test[:, 1], local_combo_array[:, 1])
        assert np.allclose(test[:, 2], local_combo_array[:, 2])

        if rank == 0:
            assert local_combo_array[0, 0] == pytest.approx(range_A[0])
            assert local_combo_array[0, 1] == pytest.approx(range_B[0])
            assert local_combo_array[0, 2] == pytest.approx(range_C[0])

        if rank == num_procs - 1:
            assert local_combo_array[-1, 0] == pytest.approx(range_A[1])
            assert local_combo_array[-1, 1] == pytest.approx(range_B[1])
            assert local_combo_array[-1, 2] == pytest.approx(range_C[1])

    @pytest.mark.component
    def test_update_model_values(self, model):
        m = model
        ps = ParameterSweep()

        param_dict = dict()
        param_dict["input_a"] = LinearSample(m.fs.input["a"], None, None, None)
        param_dict["input_b"] = LinearSample(m.fs.input["b"], None, None, None)

        original_a = value(m.fs.input["a"])
        original_b = value(m.fs.input["b"])

        new_values = [1.1 * original_a, 1.1 * original_b]

        ps._update_model_values(m, param_dict, new_values)

        assert value(m.fs.input["a"]) == pytest.approx(new_values[0])
        assert value(m.fs.input["b"]) == pytest.approx(new_values[1])

    @pytest.mark.unit
    def test_aggregate_results_arr(self):
        ps = ParameterSweep()

        input_dict = {
            "outputs": {
                "fs.output[c]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.2, 0.2, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.output[d]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.75, 0.0, 0.0, 0.75, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.slack[ab_slack]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.slack[cd_slack]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
            },
            "solve_successful": [True] * 9,
            "sweep_params": {
                "fs.input[a]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.1, 0.1, 0.0, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.input[b]": {
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.25, 0.0, 0.0, 0.25, 0.0, 0.0, 0.0, 0.0]),
                },
            },
        }

        global_num_cases = len(input_dict["sweep_params"]["fs.input[a]"]["value"])
        global_results_arr = ps._aggregate_results_arr(input_dict, global_num_cases)

        reference_results_arr = np.array(
            [
                [0.2, 0.0, 0.0, 0.0],
                [0.2, 0.75, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0],
                [1.0, 0.75, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ]
        )

        assert np.array_equal(global_results_arr, reference_results_arr)

    @pytest.mark.unit
    def test_create_local_output_skeleton(self, model):
        ps = ParameterSweep()

        m = model
        m.fs.slack_penalty = 1000.0
        m.fs.slack.setub(0)

        sweep_params = {
            "input_a": (m.fs.input["a"], 0.1, 0.9, 3),
            "input_b": (m.fs.input["b"], 0.0, 0.5, 3),
        }
        # outputs = {
        #     "output_c": m.fs.output["c"],
        #     "output_d": m.fs.output["d"],
        #     "performance": m.fs.performance,
        # }

        sweep_params, sampling_type = ps._process_sweep_params(sweep_params)
        values = ps._build_combinations(sweep_params, sampling_type, None)
        num_cases = np.shape(values)[0]
        output_dict = ps._create_local_output_skeleton(
            model, sweep_params, None, num_cases
        )

        truth_dict = {
            "outputs": {
                "input_a": {
                    "full_name": "fs.input[a]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "input_b": {
                    "full_name": "fs.input[b]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.output[c]": {
                    "full_name": "fs.output[c]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.output[d]": {
                    "full_name": "fs.output[d]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.performance": {
                    "full_name": "fs.performance",
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.slack[ab_slack]": {
                    "full_name": "fs.slack[ab_slack]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 0,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.slack[cd_slack]": {
                    "full_name": "fs.slack[cd_slack]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 0,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "fs.slack_penalty": {
                    "full_name": "fs.slack_penalty",
                    "units": "None",
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "objective": {
                    "full_name": "objective",
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
            },
            "sweep_params": {
                "input_a": {
                    "full_name": "fs.input[a]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
                "input_b": {
                    "full_name": "fs.input[b]",
                    "lower bound": 0,
                    "units": "None",
                    "upper bound": 1,
                    "value": np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                },
            },
        }

        _assert_dictionary_correctness(truth_dict, output_dict)

    @pytest.mark.unit
    def test_create_global_output(self, model):
        ps = ParameterSweep()

        m = model
        m.fs.slack_penalty = 1000.0
        m.fs.slack.setub(0)

        global_num_cases = 2 * ps.parallel_manager.number_of_worker_processes()
        sweep_params = {
            "input_a": NormalSample(m.fs.input["a"], 0.1, 0.9, global_num_cases),
            "input_b": NormalSample(m.fs.input["b"], 0.0, 0.5, global_num_cases),
        }
        outputs = {
            "output_c": m.fs.output["c"],
            "output_d": m.fs.output["d"],
            "performance": m.fs.performance,
        }

        sweep_params, sampling_type = ps._process_sweep_params(sweep_params)

        # Get the globale sweep param values
        global_values = ps._build_combinations(
            sweep_params, sampling_type, global_num_cases
        )
        # divide the workload between processors
        local_values = ps._divide_combinations(global_values)
        local_num_cases = np.shape(local_values)[0]

        local_output_dict = ps._create_local_output_skeleton(
            model, sweep_params, None, local_num_cases
        )

        # Manually update the values in the numpy array
        for key, value in local_output_dict.items():
            for subkey, subvalue in value.items():
                subvalue["value"][:] = ps.parallel_manager.get_rank()

        # Local output dict also contains the solve_successful. The solve status is
        # based on the
        local_output_dict["solve_successful"] = [True] * local_num_cases

        # Get the global output dictionary, This is properly created only on rank 0
        global_output_dict = ps._create_global_output(
            local_output_dict, global_num_cases
        )

        if ps.parallel_manager.get_rank() > 0:
            assert global_output_dict == local_output_dict
        else:
            test_array = np.repeat(
                np.arange(
                    0, ps.parallel_manager.number_of_worker_processes(), dtype=float
                ),
                2,
            )
            test_list = [True] * global_num_cases
            for key, value in global_output_dict.items():
                if key != "solve_successful":
                    for subkey, subvalue in value.items():
                        assert np.allclose(subvalue["value"], test_array)
                elif key == "solve_successful":
                    assert list(value) == test_list

    @pytest.mark.component
    @pytest.mark.parametrize("number_of_subprocesses", [1, 2, 3])
    def test_parameter_sweep(self, model, tmp_path, number_of_subprocesses):
        comm = MPI.COMM_WORLD

        tmp_path = _get_rank0_path(comm, tmp_path)

        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        ps = ParameterSweep(
            optimize_function=_optimization,
            csv_results_file_name=csv_results_file_name,
            h5_results_file_name=h5_results_file_name,
            debugging_data_dir=tmp_path,
            interpolate_nan_outputs=True,
            number_of_subprocesses=number_of_subprocesses,
        )

        # Call the parameter_sweep function
        _ = ps.parameter_sweep(
            build_model_for_tps,
            build_sweep_params_for_tps,
            build_outputs=build_outputs_for_tps,
            build_outputs_kwargs={
                "output_keys": {
                    "output_c": "fs.output[c]",
                    "output_d": "fs.output[d]",
                    "performance": "fs.performance",
                }
            },
        )

        # NOTE: rank 0 "owns" tmp_path, so it needs to be
        #       responsible for doing any output file checking
        #       tmp_path can be deleted as soon as this method
        #       returns
        if ps.parallel_manager.is_root_process():
            # Check that the global results file is created
            assert os.path.isfile(csv_results_file_name)
            assert os.path.isfile(
                os.path.join(tmp_path, "interpolated_global_results.csv")
            )

            # Check that all local output files have been created
            for k in range(ps.parallel_manager.number_of_worker_processes()):
                assert os.path.isfile(
                    os.path.join(tmp_path, f"local_results_{k:03}.h5")
                )
                assert os.path.isfile(
                    os.path.join(tmp_path, f"local_results_{k:03}.csv")
                )

            # Attempt to read in the data
            data = np.genfromtxt(csv_results_file_name, skip_header=1, delimiter=",")

            # Compare the last row of the imported data to truth
            truth_data = [0.9, 0.5, np.nan, np.nan, np.nan]
            assert np.allclose(data[-1], truth_data, equal_nan=True)

        # Check for the h5 output
        if ps.parallel_manager.is_root_process():
            truth_dict = {
                "outputs": {
                    "output_c": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.2, 0.2, np.nan, 1.0, 1.0, np.nan, np.nan, np.nan, np.nan]
                        ),
                    },
                    "output_d": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [
                                0.0,
                                0.75,
                                np.nan,
                                0.0,
                                0.75,
                                np.nan,
                                np.nan,
                                np.nan,
                                np.nan,
                            ]
                        ),
                    },
                    "performance": {
                        "value": np.array(
                            [
                                0.2,
                                0.95,
                                np.nan,
                                1.0,
                                1.75,
                                np.nan,
                                np.nan,
                                np.nan,
                                np.nan,
                            ]
                        )
                    },
                },
                "solve_successful": [
                    True,
                    True,
                    False,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                "sweep_params": {
                    "fs.input[a]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                },
            }

            read_dict = _read_output_h5(h5_results_file_name)
            _assert_dictionary_correctness(truth_dict, read_dict)
            _assert_h5_csv_agreement(csv_results_file_name, read_dict)

            # Check if there is a text file created
            import ast

            truth_txt_dict = {
                "outputs": [
                    "output_c",
                    "output_d",
                    "performance",
                ],
                "sweep_params": ["fs.input[a]", "fs.input[b]"],
            }

            txt_fpath = h5_results_file_name + ".txt"
            assert os.path.exists(txt_fpath)
            f = open(txt_fpath)
            f_contents = f.read()
            read_txt_dict = ast.literal_eval(f_contents)
            assert read_txt_dict == truth_txt_dict

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_parameter_sweep_optimize(self, model, tmp_path):
        comm = MPI.COMM_WORLD

        tmp_path = _get_rank0_path(comm, tmp_path)
        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        ps = ParameterSweep(
            optimize_function=_optimization,
            optimize_kwargs={"relax_feasibility": True},
            probe_function=_good_test_function,
            csv_results_file_name=csv_results_file_name,
            h5_results_file_name=h5_results_file_name,
            debugging_data_dir=tmp_path,
            interpolate_nan_outputs=True,
        )

        m = model
        m.fs.slack_penalty = 1000.0
        m.fs.slack.setub(0)

        A = m.fs.input["a"]
        B = m.fs.input["b"]
        sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}
        outputs = {
            "output_c": m.fs.output["c"],
            "output_d": m.fs.output["d"],
            "performance": m.fs.performance,
            "objective": m.objective,
        }
        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        # Call the parameter_sweep function
        ps.parameter_sweep(
            m,
            sweep_params,
            build_outputs=outputs,
        )

        # NOTE: rank 0 "owns" tmp_path, so it needs to be
        #       responsible for doing any output file checking
        #       tmp_path can be deleted as soon as this method
        #       returns
        if ps.parallel_manager.is_root_process():
            # Check that the global results file is created
            assert os.path.isfile(csv_results_file_name)

            # Attempt to read in the data
            data = np.genfromtxt(csv_results_file_name, skip_header=1, delimiter=",")
            # Compare the last row of the imported data to truth
            truth_data = [
                0.9,
                0.5,
                1.0,
                1.0,
                2.0,
                2.0 - 1000.0 * ((2.0 * 0.9 - 1.0) + (3.0 * 0.5 - 1.0)),
            ]
            assert np.allclose(data[-1], truth_data, equal_nan=True)

        # Check the h5
        if ps.parallel_manager.is_root_process():
            truth_dict = {
                "outputs": {
                    "output_c": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.2, 0.2, 0.2, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
                        ),
                    },
                    "output_d": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [
                                9.98580690e-09,
                                0.75,
                                1.0,
                                9.99872731e-09,
                                0.75,
                                1.0,
                                9.99860382e-09,
                                0.75,
                                1.0,
                            ]
                        ),
                    },
                    "performance": {
                        "value": np.array(
                            [0.2, 0.95, 1.2, 1.0, 1.75, 2.0, 1.0, 1.75, 2.0]
                        )
                    },
                    "objective": {
                        "value": np.array(
                            [
                                0.2,
                                9.50000020e-01,
                                -4.98799990e02,
                                1.0,
                                1.75,
                                -4.97999990e02,
                                -7.98999990e02,
                                -7.98249990e02,
                                2.0 - 1000.0 * ((2.0 * 0.9 - 1.0) + (3.0 * 0.5 - 1.0)),
                            ]
                        )
                    },
                },
                "solve_successful": [True] * 9,
                "sweep_params": {
                    "fs.input[a]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                },
            }

            read_dict = _read_output_h5(h5_results_file_name)
            _assert_dictionary_correctness(truth_dict, read_dict)
            _assert_h5_csv_agreement(csv_results_file_name, read_dict)

    @pytest.mark.component
    def test_parameter_sweep_optimize_with_added_var(self, model, tmp_path):
        # this will run a solve function that adds a variable but only in some
        # of the solves.
        """THIS TEST IS DESIGNED FOR 2 Parallel workers!!!!!!"""
        comm = MPI.COMM_WORLD

        tmp_path = _get_rank0_path(comm, tmp_path)
        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        ps = ParameterSweep(
            optimize_function=_optimization,
            initialize_function=_initialize_with_added_var,
            update_sweep_params_before_init=True,
            initialize_before_sweep=True,
            optimize_kwargs={"relax_feasibility": True},
            probe_function=_good_test_function,
            csv_results_file_name=csv_results_file_name,
            h5_results_file_name=h5_results_file_name,
            debugging_data_dir=tmp_path,
            interpolate_nan_outputs=False,
            number_of_subprocesses=2,
            parallel_back_end="MultiProcessing",
        )

        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        # Call the parameter_sweep function
        ps.parameter_sweep(
            build_model_for_tps,
            build_sweep_params_for_tps,
        )

        # NOTE: rank 0 "owns" tmp_path, so it needs to be
        #       responsible for doing any output file checking
        #       tmp_path can be deleted as soon as this method
        #       returns
        # Check that test var array was created
        if ps.parallel_manager.is_root_process():
            truth_dict = {
                "outputs": {
                    "fs.test_var": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 10,
                        "value": np.array(
                            [
                                np.nan,
                                np.nan,
                                np.nan,
                                np.nan,
                                np.nan,
                                5,
                                np.nan,
                                np.nan,
                                np.nan,
                            ]
                        ),
                    },
                    "objective": {
                        "value": np.array(
                            [
                                0.2,
                                9.50000020e-01,
                                -4.98799990e02,
                                1.0,
                                1.75,
                                -4.97999990e02,
                                -7.98999990e02,
                                -7.98249990e02,
                                2.0 - 1000.0 * ((2.0 * 0.9 - 1.0) + (3.0 * 0.5 - 1.0)),
                            ]
                        )
                    },
                },
                "solve_successful": [True] * 9,
                "sweep_params": {
                    "fs.input[a]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                },
            }

            read_dict = _read_output_h5(h5_results_file_name)
            _assert_dictionary_correctness(truth_dict, read_dict, rtol=1e-2)

    @pytest.mark.component
    def test_parameter_sweep_bad_initialize_call_2(self, model, tmp_path):
        comm = MPI.COMM_WORLD

        tmp_path = _get_rank0_path(comm, tmp_path)
        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        ps = ParameterSweep(
            optimize_function=_optimization,
            initialize_function=_reinitialize,
            initialize_kwargs={"slack_penalty": 10.0, "foo": "bar"},
            csv_results_file_name=csv_results_file_name,
            h5_results_file_name=h5_results_file_name,
            debugging_data_dir=tmp_path,
            interpolate_nan_outputs=True,
        )

        m = model
        m.fs.slack_penalty = 1000.0
        m.fs.slack.setub(0)

        A = m.fs.input["a"]
        B = m.fs.input["b"]
        sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}

        with pytest.raises(TypeError):
            # Call the parameter_sweep function
            _ = ps.parameter_sweep(
                m,
                sweep_params,
                build_outputs=None,
            )

    @pytest.mark.component
    def test_parameter_sweep_recover(self, model, tmp_path):
        comm = MPI.COMM_WORLD

        tmp_path = _get_rank0_path(comm, tmp_path)
        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        ps = ParameterSweep(
            optimize_function=_optimization,
            initialize_function=_reinitialize,
            initialize_kwargs={"slack_penalty": 10.0},
            csv_results_file_name=csv_results_file_name,
            h5_results_file_name=h5_results_file_name,
            debugging_data_dir=tmp_path,
            interpolate_nan_outputs=True,
        )
        # Call the parameter_sweep function
        _ = ps.parameter_sweep(
            build_model_for_tps,
            build_sweep_params_for_tps,
            build_outputs=None,
        )

        # NOTE: rank 0 "owns" tmp_path, so it needs to be
        #       responsible for doing any output file checking
        #       tmp_path can be deleted as soon as this method
        #       returns
        if ps.parallel_manager.is_root_process():
            # Check that the global results file is created
            assert os.path.isfile(csv_results_file_name)

            # Attempt to read in the data
            data = np.genfromtxt(csv_results_file_name, skip_header=1, delimiter=",")
            # Compare the last row of the imported data to truth
            truth_data = [0.9, 0.5, -11.0, 0.9, 0.5, 1.0, 1.0, 0.8, 0.5, 10.0, 2.0]
            assert np.allclose(data[-1], truth_data, equal_nan=True)

            # H5 dictionary test
            truth_dict = {
                "outputs": {
                    "fs.input[a]": {
                        "full_name": "fs.input[a]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "full_name": "fs.input[b]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                    "fs.output[c]": {
                        "full_name": "fs.output[c]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [
                                0.2,
                                0.2,
                                0.2,
                                1.0,
                                1.0,
                                1.0,
                                1.0,
                                1.0,
                                1.0,
                            ]
                        ),
                    },
                    "fs.output[d]": {
                        "full_name": "fs.output[d]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [
                                0.0,
                                7.50000000e-01,
                                1.0,
                                9.77756334e-09,
                                7.50000010e-01,
                                1.0,
                                9.98605188e-09,
                                7.50000010e-01,
                                1.0,
                            ]
                        ),
                    },
                    "fs.performance": {
                        "full_name": "fs.performance",
                        "value": np.array(
                            [
                                0.2,
                                0.95,
                                1.2,
                                1.0,
                                1.75,
                                2.0,
                                1.0,
                                1.75,
                                2.0,
                            ]
                        ),
                    },
                    # removed upper bound tests, as in pracice
                    # user should not be chanaing upper and lower bounds
                    # during reinitilization and current frame work
                    # this bounds get updated based on existing model in each worker
                    "fs.slack[ab_slack]": {
                        "full_name": "fs.slack[ab_slack]",
                        "lower bound": 0,
                        "units": "None",
                        "value": np.array(
                            [
                                0.0,
                                0.0,
                                -9.77208577e-09,
                                -9.54438119e-09,
                                -9.54432513e-09,
                                -9.54432505e-09,
                                0.8,
                                0.8,
                                0.8,
                            ]
                        ),
                    },
                    "fs.slack[cd_slack]": {
                        "full_name": "fs.slack[cd_slack]",
                        "lower bound": 0,
                        "units": "None",
                        "value": np.array(
                            [
                                0.0,
                                0.0,
                                0.5,
                                -9.77756334e-09,
                                -9.77226571e-09,
                                0.5,
                                -9.98605188e-09,
                                -9.77226571e-09,
                                0.5,
                            ]
                        ),
                    },
                    "fs.slack_penalty": {
                        "full_name": "fs.slack_penalty",
                        "units": "None",
                        "value": np.array(
                            [10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0]
                        ),
                    },
                    "objective": {
                        "full_name": "objective",
                        "value": np.array(
                            [
                                0.2,
                                0.95,
                                -3.79999979,
                                1.0,
                                1.75,
                                -2.99999979,
                                -6.99999978,
                                -6.24999979,
                                -10.99999979,
                            ]
                        ),
                    },
                },
                "solve_successful": [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ],
                "sweep_params": {
                    "fs.input[a]": {
                        "full_name": "fs.input[a]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "full_name": "fs.input[b]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                },
            }

            read_dict = _read_output_h5(h5_results_file_name)
            _assert_dictionary_correctness(truth_dict, read_dict)
            _assert_h5_csv_agreement(csv_results_file_name, read_dict)

    @pytest.mark.component
    def test_parameter_sweep_bad_recover(self, model, tmp_path):
        comm = MPI.COMM_WORLD

        tmp_path = _get_rank0_path(comm, tmp_path)
        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        ps = ParameterSweep(
            optimize_function=_optimization,
            initialize_function=_bad_reinitialize,
            initialize_kwargs={"slack_penalty": 10.0},
            csv_results_file_name=csv_results_file_name,
            h5_results_file_name=h5_results_file_name,
            debugging_data_dir=tmp_path,
            interpolate_nan_outputs=True,
        )
        ps.config.log_model_states = True
        # Call the parameter_sweep function
        _ = ps.parameter_sweep(
            build_model_for_tps,
            build_sweep_params_for_tps,
            build_outputs=None,
        )

        # NOTE: rank 0 "owns" tmp_path, so it needs to be
        #       responsible for doing any output file checking
        #       tmp_path can be deleted as soon as this method
        #       returns
        if ps.parallel_manager.is_root_process():
            # Check that the global results file is created
            assert os.path.isfile(csv_results_file_name)

            # Attempt to read in the data
            data = np.genfromtxt(csv_results_file_name, skip_header=1, delimiter=",")

            # Compare the last row of the imported data to truth
            truth_data = [
                0.9,
                0.5,
                np.nan,
                0.9,
                0.5,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
            ]
            assert np.allclose(data[-1], truth_data, equal_nan=True)
            # H5 dictionary test
            truth_dict = {
                "outputs": {
                    "fs.input[a]": {
                        "full_name": "fs.input[a]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "full_name": "fs.input[b]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                    "fs.output[c]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.2, 0.2, np.nan, 1.0, 1.0, np.nan, np.nan, np.nan, np.nan]
                        ),
                    },
                    "fs.output[d]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [
                                0.0,
                                0.75,
                                np.nan,
                                0.0,
                                0.75,
                                np.nan,
                                np.nan,
                                np.nan,
                                np.nan,
                            ]
                        ),
                    },
                    "fs.performance": {
                        "value": np.array(
                            [
                                0.2,
                                0.95,
                                np.nan,
                                1.0,
                                1.75,
                                np.nan,
                                np.nan,
                                np.nan,
                                np.nan,
                            ]
                        )
                    },
                    "fs.slack[ab_slack]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 0,
                        "value": np.array(
                            [0.0, 0.0, np.nan, 0.0, 0.0, np.nan, np.nan, np.nan, np.nan]
                        ),
                    },
                    "fs.slack[cd_slack]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 0,
                        "value": np.array(
                            [0.0, 0.0, np.nan, 0.0, 0.0, np.nan, np.nan, np.nan, np.nan]
                        ),
                    },
                    "objective": {
                        "value": np.array(
                            [
                                0.2,
                                0.95,
                                np.nan,
                                1.0,
                                1.75,
                                np.nan,
                                np.nan,
                                np.nan,
                                np.nan,
                            ]
                        )
                    },
                },
                "solve_successful": [
                    True,
                    True,
                    False,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                "sweep_params": {
                    "fs.input[a]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                },
            }
            true_init_state, true_solved_state = dummy_kernel_logic(
                truth_dict["solve_successful"]
            )
            if has_mpi_peer_processes == False:
                assert true_init_state == ps.model_manager.initialized_states["state"]
                assert true_solved_state == ps.model_manager.solved_states["state"]
            read_dict = _read_output_h5(h5_results_file_name)
            _assert_dictionary_correctness(truth_dict, read_dict)
            _assert_h5_csv_agreement(csv_results_file_name, read_dict)

    @pytest.mark.component
    def test_parameter_sweep_force_initialize(self, model, tmp_path):
        results_fname = os.path.join(tmp_path, "global_results_force_initialize")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"
        ps = ParameterSweep(
            optimize_function=_optimization,
            initialize_before_sweep=True,
            initialize_function=_reinitialize,
            initialize_kwargs={"slack_penalty": 10.0},
            csv_results_file_name=csv_results_file_name,
            h5_results_file_name=h5_results_file_name,
            debugging_data_dir=tmp_path,
            interpolate_nan_outputs=False,
        )

        # Call the parameter_sweep function
        _ = ps.parameter_sweep(
            build_model_for_tps,
            build_sweep_params_for_tps,
            build_outputs=None,
        )

        # NOTE: rank 0 "owns" tmp_path, so it needs to be
        #       responsible for doing any output file checking
        #       tmp_path can be deleted as soon as this method
        #       returns
        if ps.parallel_manager.is_root_process():
            # Check that the global results file is created
            assert os.path.isfile(csv_results_file_name)

            # Attempt to read in the data
            data = np.genfromtxt(csv_results_file_name, skip_header=1, delimiter=",")

            # Compare the last row of the imported data to truth
            truth_data = [0.9, 0.5, -11.0, 0.9, 0.5, 1.0, 1.0, 0.8, 0.5, 10.0, 2.0]
            assert np.allclose(data[-1], truth_data, equal_nan=True)

            # H5 dictionary test
            truth_dict = {
                "outputs": {
                    "fs.input[a]": {
                        "full_name": "fs.input[a]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "full_name": "fs.input[b]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                    "fs.output[c]": {
                        "full_name": "fs.output[c]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [
                                0.2,
                                0.2,
                                0.2,
                                1.0,
                                1.0,
                                1.0,
                                1.0,
                                1.0,
                                1.0,
                            ]
                        ),
                    },
                    "fs.output[d]": {
                        "full_name": "fs.output[d]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [
                                9.77899282e-09,
                                7.50000010e-01,
                                1.00000001e00,
                                9.77868510e-09,
                                7.50000010e-01,
                                1.00000001e00,
                                9.90100346e-09,
                                7.50000010e-01,
                                1.00000001e00,
                            ]
                        ),
                    },
                    "fs.performance": {
                        "full_name": "fs.performance",
                        "value": np.array(
                            [
                                0.2,
                                0.95,
                                1.2,
                                1.00,
                                1.75,
                                2.0,
                                1.0,
                                1.75,
                                2.0,
                            ]
                        ),
                    },
                    "fs.slack[ab_slack]": {
                        "full_name": "fs.slack[ab_slack]",
                        "lower bound": 0,
                        "units": "None",
                        # "upper bound": 0,
                        "value": np.array(
                            [
                                -9.77219059e-09,
                                -9.77208576e-09,
                                -9.77208577e-09,
                                -9.54438113e-09,
                                -9.54432509e-09,
                                -9.54432508e-09,
                                7.99999990e-01,
                                7.99999990e-01,
                                7.99999990e-01,
                            ]
                        ),
                    },
                    "fs.slack[cd_slack]": {
                        "full_name": "fs.slack[cd_slack]",
                        "lower bound": 0,
                        "units": "None",
                        # "upper bound": 0,
                        "value": np.array(
                            [
                                -9.77899282e-09,
                                -9.77226572e-09,
                                4.99999990e-01,
                                -9.77868510e-09,
                                -9.77226572e-09,
                                4.99999990e-01,
                                -9.90100346e-09,
                                -9.77226572e-09,
                                4.99999990e-01,
                            ]
                        ),
                    },
                    "fs.slack_penalty": {
                        "full_name": "fs.slack_penalty",
                        "units": "None",
                        "value": np.array(
                            [10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0]
                        ),
                    },
                    "objective": {
                        "full_name": "objective",
                        "value": np.array(
                            [
                                0.20000022,
                                0.95000021,
                                -3.79999979,
                                1.00000021,
                                1.75000021,
                                -2.99999979,
                                -6.99999978,
                                -6.24999979,
                                -10.99999979,
                            ]
                        ),
                    },
                },
                "solve_successful": [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ],
                "sweep_params": {
                    "fs.input[a]": {
                        "full_name": "fs.input[a]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "full_name": "fs.input[b]",
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                },
            }

            read_dict = _read_output_h5(h5_results_file_name)
            _assert_dictionary_correctness(truth_dict, read_dict)
            _assert_h5_csv_agreement(csv_results_file_name, read_dict)

    @pytest.mark.component
    def test_parameter_sweep_bad_sweep_update_before_initialize(self, model, tmp_path):
        ps = ParameterSweep(
            optimize_function=_optimization,
            initialize_before_sweep=True,
            initialize_function=None,
            initialize_kwargs=None,
            update_sweep_params_before_init=True,
        )
        m = model
        m.fs.slack_penalty = 1000.0
        m.fs.slack.setub(0)

        A = m.fs.input["a"]
        B = m.fs.input["b"]
        sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}

        with pytest.raises(ValueError):
            # Call the parameter_sweep function
            ps.parameter_sweep(
                m,
                sweep_params,
                build_outputs=None,
            )

    @pytest.mark.component
    def test_parameter_sweep_bad_optimization_call(self, model, tmp_path):
        ps = ParameterSweep(
            optimize_function=_optimization,
            optimize_kwargs={"foo": "bar"},
        )

        m = model
        m.fs.slack_penalty = 1000.0
        m.fs.slack.setub(0)

        A = m.fs.input["a"]
        B = m.fs.input["b"]
        sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}

        with pytest.raises(TypeError):
            # Call the parameter_sweep function
            ps.parameter_sweep(
                m,
                sweep_params,
                build_outputs=None,
            )

    @pytest.mark.component
    def test_parameter_sweep_bad_reinitialize_call(self, model, tmp_path):
        def reinit(a=42):
            pass

        ps = ParameterSweep(
            optimize_function=_optimization,
            initialize_before_sweep=True,
            initialize_function=reinit,
            initialize_kwargs={"foo": "bar"},
        )

        m = model
        m.fs.slack_penalty = 1000.0
        m.fs.slack.setub(0)

        A = m.fs.input["a"]
        B = m.fs.input["b"]
        sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}

        with pytest.raises(TypeError):
            # Call the parameter_sweep function
            ps.parameter_sweep(
                m,
                sweep_params,
                build_outputs=None,
            )

    @pytest.mark.component
    def test_parameter_sweep_probe_fail(self, model, tmp_path):
        comm = MPI.COMM_WORLD

        tmp_path = _get_rank0_path(comm, tmp_path)
        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        ps = ParameterSweep(
            optimize_function=_optimization,
            optimize_kwargs={"relax_feasibility": True},
            probe_function=_bad_test_function,
            csv_results_file_name=csv_results_file_name,
            h5_results_file_name=h5_results_file_name,
            debugging_data_dir=tmp_path,
            interpolate_nan_outputs=True,
        )

        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        # Call the parameter_sweep function
        ps.parameter_sweep(
            build_model_for_tps,
            build_sweep_params_for_tps,
            build_outputs=build_outputs_for_tps,
            build_outputs_kwargs={
                "output_keys": {
                    "output_c": "fs.output[c]",
                    "output_d": "fs.output[d]",
                    "performance": "fs.performance",
                    "objective": "objective",
                }
            },
        )

        # NOTE: rank 0 "owns" tmp_path, so it needs to be
        #       responsible for doing any output file checking
        #       tmp_path can be deleted as soon as this method
        #       returns
        if ps.parallel_manager.is_root_process():
            # Check that the global results file is created
            assert os.path.isfile(csv_results_file_name)

            # Attempt to read in the data
            data = np.genfromtxt(csv_results_file_name, skip_header=1, delimiter=",")
            # Compare the last row of the imported data to truth
            truth_data = [
                0.9,
                0.5,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
            ]
            assert np.allclose(data[-1], truth_data, equal_nan=True)

            truth_dict = {
                "outputs": {
                    "output_c": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array([np.nan] * 9),
                    },
                    "output_d": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            np.array([np.nan] * 9),
                        ),
                    },
                    "performance": {"value": np.array([np.nan] * 9)},
                    "objective": {
                        "value": np.array(
                            np.array([np.nan] * 9),
                        )
                    },
                },
                "solve_successful": [False] * 9,
                "sweep_params": {
                    "fs.input[a]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                },
            }

            read_dict = _read_output_h5(h5_results_file_name)
            _assert_dictionary_correctness(truth_dict, read_dict)
            _assert_h5_csv_agreement(csv_results_file_name, read_dict)

    @pytest.mark.component
    def test_parameter_sweep_function(self, model, tmp_path):
        comm = MPI.COMM_WORLD
        rank = comm.rank
        tmp_path = _get_rank0_path(comm, tmp_path)

        m = model
        m.fs.slack_penalty = 1000.0
        m.fs.slack.setub(0)

        A = m.fs.input["a"]
        B = m.fs.input["b"]
        sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}
        outputs = {
            "output_c": m.fs.output["c"],
            "output_d": m.fs.output["d"],
            "performance": m.fs.performance,
        }

        results_fname = os.path.join(tmp_path, "global_results")
        csv_results_file_name = str(results_fname) + ".csv"
        h5_results_file_name = str(results_fname) + ".h5"

        # Call the parameter_sweep function
        parameter_sweep(
            m,
            sweep_params,
            build_outputs=outputs,
            csv_results_file_name=csv_results_file_name,
            h5_results_file_name=h5_results_file_name,
            optimize_function=_optimization,
            debugging_data_dir=tmp_path,
            interpolate_nan_outputs=True,
        )

        # NOTE: rank 0 "owns" tmp_path, so it needs to be
        #       responsible for doing any output file checking
        #       tmp_path can be deleted as soon as this method
        #       returns
        if rank == 0:
            # Check that the global results file is created
            assert os.path.isfile(csv_results_file_name)
            assert os.path.isfile(
                os.path.join(tmp_path, "interpolated_global_results.csv")
            )

            # Check that all local output files have been created
            for k in range(comm.size):
                assert os.path.isfile(
                    os.path.join(tmp_path, f"local_results_{k:03}.h5")
                )
                assert os.path.isfile(
                    os.path.join(tmp_path, f"local_results_{k:03}.csv")
                )

            # Attempt to read in the data
            data = np.genfromtxt(csv_results_file_name, skip_header=1, delimiter=",")

            # Compare the last row of the imported data to truth
            truth_data = [0.9, 0.5, np.nan, np.nan, np.nan]
            assert np.allclose(data[-1], truth_data, equal_nan=True)

        # Check for the h5 output
        if rank == 0:
            truth_dict = {
                "outputs": {
                    "output_c": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.2, 0.2, np.nan, 1.0, 1.0, np.nan, np.nan, np.nan, np.nan]
                        ),
                    },
                    "output_d": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [
                                0.0,
                                0.75,
                                np.nan,
                                0.0,
                                0.75,
                                np.nan,
                                np.nan,
                                np.nan,
                                np.nan,
                            ]
                        ),
                    },
                    "performance": {
                        "value": np.array(
                            [
                                0.2,
                                0.95,
                                np.nan,
                                1.0,
                                1.75,
                                np.nan,
                                np.nan,
                                np.nan,
                                np.nan,
                            ]
                        )
                    },
                },
                "solve_successful": [
                    True,
                    True,
                    False,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                "sweep_params": {
                    "fs.input[a]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.1, 0.1, 0.1, 0.5, 0.5, 0.5, 0.9, 0.9, 0.9]
                        ),
                    },
                    "fs.input[b]": {
                        "lower bound": 0,
                        "units": "None",
                        "upper bound": 1,
                        "value": np.array(
                            [0.0, 0.25, 0.5, 0.0, 0.25, 0.5, 0.0, 0.25, 0.5]
                        ),
                    },
                },
            }

            read_dict = _read_output_h5(h5_results_file_name)
            _assert_dictionary_correctness(truth_dict, read_dict)
            _assert_h5_csv_agreement(csv_results_file_name, read_dict)

            # Check if there is a text file created
            import ast

            truth_txt_dict = {
                "outputs": [
                    "output_c",
                    "output_d",
                    "performance",
                ],
                "sweep_params": ["fs.input[a]", "fs.input[b]"],
            }

            txt_fpath = h5_results_file_name + ".txt"
            assert os.path.exists(txt_fpath)
            f = open(txt_fpath)
            f_contents = f.read()
            read_txt_dict = ast.literal_eval(f_contents)
            assert read_txt_dict == truth_txt_dict

    @pytest.mark.unit
    def test_parameter_sweep_custom_do_param_sweep(self, model, tmp_path):
        def custom_do_param_sweep(model, sweep_params, outputs, local_values, **kwargs):
            return kwargs

        custom_kwargs = {"val1": 2.0}
        ps = ParameterSweep(
            custom_do_param_sweep=custom_do_param_sweep,
            custom_do_param_sweep_kwargs=custom_kwargs,
        )

        m = model
        m.fs.slack_penalty = 1000.0
        m.fs.slack.setub(0)

        A = m.fs.input["a"]
        B = m.fs.input["b"]
        sweep_params = {A.name: (A, 0.1, 0.9, 3), B.name: (B, 0.0, 0.5, 3)}
        outputs = {
            "output_c": m.fs.output["c"],
            "output_d": m.fs.output["d"],
            "performance": m.fs.performance,
            "objective": m.objective,
        }

        assert ps.config.custom_do_param_sweep is not None
        return_dict = ps.config.custom_do_param_sweep(
            model,
            sweep_params,
            outputs,
            0.0,
            val1=2.0,
        )
        assert return_dict == custom_kwargs


def _optimization(m, relax_feasibility=False):
    if relax_feasibility:
        m.fs.slack.setub(None)

    solver = pyo.SolverFactory("ipopt")
    results = solver.solve(m)

    return results


def _initialize_with_added_var(m):
    if (abs(m.fs.input["a"].value - 0.5) < 1e-6) and abs(
        m.fs.input["b"].value - 0.5
    ) < 1e-6:
        m.fs.test_var = pyo.Var(initialize=5, bounds=(0, 10))
        m.fs.test_var.fix()


def _reinitialize(m, slack_penalty=10.0):
    m.fs.slack.setub(None)
    m.fs.slack_penalty.value = slack_penalty


def _bad_reinitialize(m, **kwargs):
    pass


def _get_rank0_path(comm, tmp_path):
    if comm is None:
        return tmp_path
    return comm.bcast(tmp_path, root=0)


def _good_test_function(m):
    return True


def _bad_test_function(m):
    return False


def _assert_dictionary_correctness(truth_dict, test_dict, rtol=1e-05, atol=1e-08):
    assert truth_dict.keys() == test_dict.keys()

    for key, item in truth_dict.items():
        if key in ["sweep_params", "outputs"]:
            for subkey, subitem in item.items():
                for subsubkey, subsubitem in subitem.items():
                    if subsubkey == "value":
                        assert np.allclose(
                            test_dict[key][subkey]["value"],
                            subitem["value"],
                            equal_nan=True,
                            rtol=rtol,
                            atol=atol,
                        )
                    else:
                        if (
                            subsubkey in ["upper bound", "lower bound"]
                            and test_dict[key][subkey][subsubkey] is None
                        ):
                            assert subsubitem in [np.finfo("d").min, np.finfo("d").max]
                        else:
                            assert subsubitem == test_dict[key][subkey][subsubkey]
        elif key == "solve_successful":
            assert item == test_dict[key]
        elif key in ["nominal_idx", "differential_idx"]:
            assert np.allclose(
                test_dict[key], item, equal_nan=True, rtol=rtol, atol=atol
            )


def _assert_h5_csv_agreement(csv_filename, h5_dict):
    csv_header = _build_header_list_from_csv(csv_filename)
    csv_data = np.genfromtxt(csv_filename, skip_header=1, delimiter=",")

    for output_type in ["sweep_params", "outputs"]:
        for key in h5_dict[output_type].keys():
            # Find the matching label in the CSV file
            idx = csv_header.index(key)
            assert np.allclose(
                h5_dict[output_type][key]["value"], csv_data[:, idx], equal_nan=True
            )


def _build_header_list_from_csv(csv_filename):
    with open(csv_filename, "r") as fp:
        csv_header_raw = fp.readline()

    csv_header_raw = csv_header_raw.replace("# ", "").strip()
    csv_header_raw = csv_header_raw.split(",")

    csv_header = []

    for k, item in enumerate(csv_header_raw):
        if k == 0:
            csv_header.append(item)

        else:
            if csv_header[-1].count("[") == csv_header[-1].count("]"):
                csv_header.append(item)
            else:
                csv_header[-1] += f",{item}"

    return csv_header


def _read_output_h5(filevar):
    if isinstance(filevar, str):
        f = h5py.File(filevar, "r")
    elif isinstance(filevar, h5py.Group):
        f = filevar
    else:
        raise RuntimeError(f"Unrecognized type for filepath {type(filevar)}")

    l1_keys = list(f.keys())
    output_dict = {}
    for key in l1_keys:  # Input or Output
        if key in ["sweep_params", "outputs"]:  #  "solve_successful":
            output_dict[key] = {}
            l2_keys = list(f[key].keys())
            for subkey in l2_keys:  # Variable name
                output_dict[key][subkey] = {}
                l3_keys = list(f[key][subkey].keys())
                for subsubkey in l3_keys:  # variable metadata
                    output_dict[key][subkey][subsubkey] = f[key][subkey][subsubkey][()]
                    if subsubkey == "units" or subsubkey == "full_name":
                        # The strings are recovered in bytes. we choose to convert it to utf-8
                        output_dict[key][subkey][subsubkey] = output_dict[key][subkey][
                            subsubkey
                        ].decode("utf-8")
        elif key == "solve_successful":
            output_dict[key] = list(f[key]["solve_successful"][()])
        elif key in ["nominal_idx", "differential_idx"]:
            output_dict[key] = f[key][key][()]

    if isinstance(filevar, str):
        f.close()

    return output_dict



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains the base class for interacting with WaterTAP data files
with zero-order model parameter data.
"""
import os
import yaml
from copy import deepcopy


class Database:
    """
    WaterTap Database class.

    Used to instantiate an instance of a database for loading parameters
    associated with zero-order models in WaterTap.

    Args:
        dbpath - (optional) path to database folder containing yaml files

    Returns:
        an instance of a Database object linked to the provided database
    """

    def __init__(self, dbpath=None):
        self._cached_files = {}

        if dbpath is None:
            self._dbpath = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "..",
                "data",
                "techno_economic",
            )
        else:
            self._dbpath = dbpath

            # Confirm valid path
            if not os.path.isdir(self._dbpath):
                raise OSError(
                    f"Could not find requested path {self._dbpath}. Please "
                    f"check that this path exists."
                )

        # Create placeholder _component_list attribute
        self._component_list = None

    def get_source_data(self, water_source=None):
        """
        Method to retrieve water source definition from database.

        Args:
            water_source - (optional) string indicating specific water source.
                           If None, the default water source will be used.

        Returns:
            dict of parameters defined in database for given water source

        Raises:
            KeyError if database has not defined water sources
        """
        if "water_sources" in self._cached_files:
            # If data is already in cached files use this
            source_data = self._cached_files["water_sources"]
        else:
            # Else load data from required file
            try:
                with open(os.path.join(self._dbpath, "water_sources.yaml"), "r") as f:
                    lines = f.read()
                    f.close()
            except OSError:
                raise KeyError("Could not find water_sources.yaml in database.")

            source_data = yaml.load(lines, yaml.Loader)

            # Store data in cache and return
            self._cached_files["water_sources"] = source_data

        # Check that water source is defined
        if water_source is None:
            try:
                water_source = source_data["default"]
            except KeyError:
                raise KeyError(
                    "Database has not defined a default water source and "
                    "none was provided."
                )

        return source_data[water_source]

    def get_solute_set(self, water_source=None):
        """
        Method to retrieve solute set for a given water source.

        Args:
            water_source - (optional) string indicating specific water source.
                           If None, the default water source will be used.

        Returns:
            list of solutes contained in the database for the given source.

        Raises:
            KeyError if water source could not be found in database
        """
        source_data = self.get_source_data(water_source)

        # Get component set for water source
        comp_set = list(source_data["solutes"].keys())

        return comp_set

    def get_unit_operation_parameters(self, technology, subtype=None):
        """
        Method to retrieve parameters for a given technology by subtype.

        Args:
            technology - unit operation technology to look up and retrieve
                         parameters for.
            subtype - (optional) string or list-of-strings indicating specific
                      sub-type of technology to return parameters for. If not
                      provided, the default parameters are used instead.

        Returns:
            dict of parameters for technology and subtype

        Raises:
            KeyError if technology or subtype could not be found in database
            TypeError if subytpe is not string or list-of-strings
        """
        params = self._get_technology(technology)

        # Get default parameter set
        sparams = deepcopy(params["default"])

        if subtype is None:
            # Return default values
            pass
        elif isinstance(subtype, str):
            try:
                sparams.update(params[subtype])
            except KeyError:
                raise KeyError(
                    f"Received unrecognised subtype {subtype} for technology "
                    f"{technology}."
                )
        else:
            # Assume subtype is list-like and raise an exception if not
            try:
                for s in subtype:
                    # Iterate through provided subtypes and update parameters
                    # Note that this will overwrite previous parameters if
                    # there is overlap, so we might need to be careful in use.
                    try:
                        sparams.update(deepcopy(params[s]))
                    except KeyError:
                        raise KeyError(
                            f"Received unrecognised subtype {s} for "
                            f"technology {technology}."
                        )
            except TypeError:
                raise TypeError(
                    f"Unexpected type for subtype {subtype}: must be string "
                    f"or list like."
                )

        return sparams

    def flush_cache(self):
        """
        Method to flush cached files in database object.
        """
        self._cached_files = {}

    @property
    def component_list(self):
        return self._return_component_list()

    def _return_component_list(self):
        if self._component_list is None:
            self._load_component_list()
        return self._component_list

    def _get_technology(self, technology):
        if technology in self._cached_files:
            # If data is already in cached files, return
            return self._cached_files[technology]
        else:
            # Else load data from required file
            try:
                with open(os.path.join(self._dbpath, technology + ".yaml"), "r") as f:
                    lines = f.read()
                    f.close()
            except OSError:
                raise KeyError(f"Could not find entry for {technology} in database.")

            fdata = yaml.load(lines, yaml.Loader)

            # Store data in cache and return
            self._cached_files[technology] = fdata
            return fdata

    def _load_component_list(self):
        """
        Load list of supported components from component_list.yaml file and
        store as _component_list attribute.

        Returns:
            None
        """
        try:
            with open(os.path.join(self._dbpath, "component_list.yaml"), "r") as f:
                lines = f.read()
                f.close()
        except OSError:
            raise KeyError("Could not find component_list.yaml in database.")

        self._component_list = yaml.load(lines, yaml.Loader)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
#################################################################################

__author__ = "Paul Vecchiarelli, Adam Atia"

from pyomo.environ import (
    units as pyunits,
    ConcreteModel,
    assert_optimal_termination,
)

from idaes.core import FlowsheetBlock, MaterialFlowBasis
from idaes.core.util.scaling import calculate_scaling_factors
from idaes.core.solvers import get_solver

from watertap.property_models.multicomp_aq_sol_prop_pack import MCASParameterBlock

from watertap.tools.oli_api.util.watertap_to_oli_helper_functions import (
    get_molar_mass,
    get_charge,
)


def create_state_block(source_water):
    """
    Creates a state block using the Multi Component Aqueous Solution (MCAS) property model.

    :param source_water: dictionary containing state variables and units

    :return m: ConcreteModel containing MCAS state block
    """

    solver = get_solver()
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    props = create_property_model_input(
        source_water["components"], property_model_type="mcas"
    )
    m.fs.properties = MCASParameterBlock(
        **props, material_flow_basis=MaterialFlowBasis.mass
    )
    stream = m.fs.stream = m.fs.properties.build_state_block([0])

    for comp, conc in stream[0].flow_mass_phase_comp.items():
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1 / conc.value, index=comp
        )
    for comp, conc in stream[0].conc_mass_phase_comp.items():
        m.fs.properties.set_default_scaling(
            "conc_mass_phase_comp", 1 / conc.value, index=comp
        )
    m.fs.properties.set_default_scaling(
        "flow_vol_phase", 1 / stream[0].flow_vol_phase["Liq"].value, index=("Liq")
    )
    calculate_scaling_factors(m)

    convert_to_state_block_units(stream[0].temperature, source_water, "temperature")
    convert_to_state_block_units(stream[0].pressure, source_water, "pressure")
    stream[0].conc_mass_phase_comp
    convert_conc = lambda conc: pyunits.convert_value(
        conc,
        from_units=source_water["units"]["components"],
        to_units=stream[0].conc_mass_phase_comp._units,
    )

    # TODO: enable customization
    conc_basis = "conc_mass_phase_comp"
    vol_basis = "flow_vol_phase"
    phase = "Liq"
    var_args = {}
    for comp in source_water["components"]:
        var_args[(conc_basis, (phase, comp))] = convert_conc(
            source_water["components"][comp]
        )
    var_args.update({(vol_basis, phase): 1e-3})
    stream.calculate_state(var_args=var_args, hold_state=True)
    stream.initialize()
    result = solver.solve(m)
    assert_optimal_termination(result)
    return m


def convert_to_state_block_units(state_variable, source: dict, key):
    """
    Converts state variable values from input source to state block units.

    :param state_variable: state block attribute (i.e. temperature, pressure, concentration) to convert
    :param source: input source with initial value
    :param key: lookup key in input source for specified state variable
    """

    converted_value = pyunits.convert_value(
        source[key], from_units=source["units"][key], to_units=state_variable._units
    )
    state_variable.fix(converted_value)


def create_property_model_input(components, property_model_type: str = ""):
    """
    Builds property package inputs.

    :param components: dictionary containing solute concentrations in mg/L
    :param property_model_type: string specifying property model to use

    :return property_model_inputs: dict containing inputs needed to build property model
    """

    if property_model_type == "mcas":
        solute_list = []
        mw_data = {}
        charge = {}

        for component in components:
            molar_mass = get_molar_mass(component)
            charge_value = get_charge(component)
            solute_list.append(component)
            mw_data[component] = get_molar_mass(component) * 1e-3
            charge_value = get_charge(component)

            if charge_value != 0:
                charge[component] = charge_value

        property_model_inputs = {
            "solute_list": solute_list,
            "mw_data": mw_data,
            "charge": charge,
        }
        return property_model_inputs


def extract_state_vars(state_block, conc_var, units):
    """
    Extracts state variables from state block into a dictionary.

    :param state_block: input source to extract values from
    :param conc_var: state block attribute containing concentrations
    :param units: dict containing PYOMO unit objects

    :return state_vars: dictionary containing state variables
    """

    update_conc = lambda conc, var, key: pyunits.convert_value(
        conc, var._units, units[key]
    )
    components = {}
    for comp, conc in conc_var.items():
        phase = comp[0]
        solute = comp[1]
        if solute != "H2O":
            components[solute] = update_conc(conc.value, conc_var, "components")
    update_t_p = lambda var, key: pyunits.convert_value(
        var.value, var._units, units[key]
    )
    temp = update_t_p(state_block.temperature, "temperature")
    pressure = update_t_p(state_block.pressure, "pressure")
    state_vars = {
        "components": components,
        "temperature": temp,
        "pressure": pressure,
    }
    return state_vars
